package art.soft;

import java.awt.Frame;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.awt.image.IndexColorModel;
import java.awt.image.MemoryImageSource;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import javax.swing.JFileChooser;

/**
 *
 * @author Артём Святоха
 */
public class Emul extends Frame
    implements Runnable, MouseListener, MouseMotionListener, KeyListener {

    private Thread thread;

    private byte ram[];
    private byte rom[];

    //private Image buffer;
    //private Graphics g;

    public Emul() {
        super("Эмулятор Macintosh Plus");
        //
        setSize(SCR_WIDTH + 24, SCR_HEIGHT + 46);
        setFocusable(true);
        //
        InputStream bios = getClass().getResourceAsStream("/res/bios.rom");
        try {
            rom = new byte[bios.available()];
            bios.read(rom, 0, rom.length);
            bios.close();
        } catch(IOException e) {}
        //
        Toolkit tk = Toolkit.getDefaultToolkit();
        Image invis = tk.createImage(new MemoryImageSource(16, 16, new int[256], 0, 16));
        setCursor(tk.createCustomCursor(invis, new Point(), null));
    }

    public void start() {
        InitKeyboardMap();
        //
        //buffer = createImage(SCR_WIDTH, SCR_HEIGHT);
        //g = buffer.getGraphics();
        //buffer = new BufferedImage(SCR_WIDTH, SCR_HEIGHT, BufferedImage.TYPE_BYTE_BINARY);
        //
        // Инициализация оперативной памяти
        ram = new byte[0x00400000]; // 4 MB RAM
        //ram = new byte[0x00100000]; // 512 KB RAM
        //
        InitVideo();
        //
        InitFileManager();
        InitDrives();
        RTC_Init();
        m68k_reset();
        //
        addMouseListener(this);
        addMouseMotionListener(this);
        addKeyListener(this);
        //
        thread = new Thread(this);
        thread.start();
        setVisible(true);
    }

    private void customreset() {
        IWM_Reset();
	SCC_Reset();
	SCSI_Reset();
	VIA_Reset();
        Memory_Reset();
	Sony_Reset();
    }

    private final JFileChooser fileManager = new JFileChooser();

    private void InitFileManager() {
        fileManager.setDialogTitle("Открыть DMG образ");
        fileManager.setCurrentDirectory(new File("."));
        fileManager.setDialogType(JFileChooser.OPEN_DIALOG);
        fileManager.setFileSelectionMode(JFileChooser.FILES_ONLY);
    }

    // Эмуляция видео чипа начало
    private boolean vDraw = false;
    private int vPage2 = 1;
    private static final int vX = 12, vY = 36; // Смещения перемещения
    
    private static final int SCR_WIDTH = 512;
    private static final int SCR_WIDTH_BYTES = 64;
    private static final int SCR_HEIGHT = 342;

    private BufferedImage vPage0, vPage1;

    private void InitVideo() {
        byte[] bw = {(byte) 0xff, (byte) 0};
        IndexColorModel blackAndWhite = new IndexColorModel(1, 2, bw, bw, bw);
        WritableRaster rPage1 = Raster.createPackedRaster(
                new DataBufferByte(ram, 0x5900, ram.length - 0x5900),
                SCR_WIDTH, SCR_HEIGHT, 1, new Point());
        WritableRaster rPage0 = Raster.createPackedRaster(
                new DataBufferByte(ram, 0x5900, ram.length - 0xD900),
                SCR_WIDTH, SCR_HEIGHT, 1, new Point());
        vPage0 = new BufferedImage(blackAndWhite, rPage0, false, null);
        vPage1 = new BufferedImage(blackAndWhite, rPage1, false, null);
    }

    private void VIA_PORA6(int Data) {
        vPage2 = Data;
    }

    @Override
    public void update(Graphics g) {
        paint(g);
    }

    @Override
    public void paint(Graphics g){
        vDraw = true;
        g.drawImage(vPage2 == 1 ? vPage1 : vPage0, vX, vY, getWidth() - 24, getHeight() - 46, this);
        vDraw = false;
    }
    // Эмуляция видео чипа конец

    // Эмуляция мыши начало
    private int CurMouseH = SCR_WIDTH / 2;
    private int CurMouseV = SCR_HEIGHT / 2;
    private boolean CurMouseButton;

    private void Mouse_Update2() {
        if (Mouse_Enabled()) { // SCC activated yet
            int NewMouse = (CurMouseV << 16) | CurMouseH;
            if (do_get_mem_long(ram, 0x0828) != NewMouse) {
                do_put_mem_long(ram, 0x0828, NewMouse); // Set Mouse Position
                do_put_mem_long(ram, 0x082C, NewMouse);
                put_long(0x0830, NewMouse);
                ram[0x08CE] = -1; // Tell MacOS to redraw the Mouse
            }
        }
    }

    private int VIA_GORB3() { // Mouse Button
        return CurMouseButton ? 0 : -1;
    }
    // Эмуляция мыши конец

    // Эмуляция клавиатуры начало
    private int CommandPending = 0;

    private void Keyboard_Put(int in) {
        CommandPending = in;
    }

    private final byte[] theKeys = new byte[16];
    private final byte[] theKeyCopys = new byte[16];
	/*
		What the emulated keyboard thinks the mac thinks is the
		state of the keyboard. This is compared to theKeys,
		which is the actual state of the real keyboard.
	*/

    private int InstantCommandData  = 0x7B;

    private boolean AttemptToFinishInquiry() {
        int i = 0;
        boolean KeyDown = false;
        int Keyboard_Data;
        //
        int j;
        int b;
        boolean FindKeyEvent = false;
        for (j = 0; j < 16; ++j) {
            int k1 = theKeys[j];
            int k2 = theKeyCopys[j];
            if (k1 != k2) {
                int bit = 1;
                for (b = 0; b < 8; ++b) {
                    int newk = k1 & bit;
                    int oldk = k2 & bit;
                    if (oldk != newk) {
                        theKeyCopys[j] = (byte) ((k2 & ~bit) | newk);
                        i = j * 8 + b;
                        KeyDown = (newk != 0);
                        FindKeyEvent = true;
                    }
                    bit <<= 1;
                }
            }
        }
        if (FindKeyEvent) {
            if (i < 64) {
                Keyboard_Data = i << 1;
                if (!KeyDown) {
                    Keyboard_Data += 128;
                }
            } else {
                Keyboard_Data = 121;
                InstantCommandData = (i - 64) << 1;
                if (!KeyDown) {
                    InstantCommandData += 128;
                }
            }
            GotKeyBoardData(Keyboard_Data);
            return true;
        } else {
            return false;
        }
    }

    private int InquiryCommandTimer = 0;

    private void Keyboard_Get() {
        if (InquiryCommandTimer != 0) {
            InquiryCommandTimer = 0; /* abort Inquiry */
        }
        switch (CommandPending) {
            case 0x10: // Inquiry Command
                if (!AttemptToFinishInquiry()) {
                    InquiryCommandTimer = 16;
                }
                break;
            case 0x14: // Instant Command
                GotKeyBoardData(InstantCommandData);
                break;
            case 0x16: // Model Command
            {
                int i;

                for (i = 4; --i >= 0;) {
                    theKeyCopys[i] = 0;
                }
            }
            GotKeyBoardData(0x0b /*0x01*/); // Test value, means Model 0, no extra devices
            //Fixed by Hoshi Takanori - it uses the proper keyboard type now
            break;
            case 0x36: // Test Command
                GotKeyBoardData(0x7D);
                break;
            case 0x00:
                GotKeyBoardData(0);
                break;
            default:
                /* Debugger(); */
                GotKeyBoardData(0);
                break;
        }
    }

    private void Keyboard_UpdateKeyMap(int key, boolean down) {
        if (key >= 0 && key < 128) {
            int bit = 1 << (key & 7);
            if (down) {
                theKeys[key >>> 3] |= bit;
            } else {
                theKeys[key >>> 3] &= ~bit;
            }
        }
    }

    private void KeyBoard_Update() {
        if (InquiryCommandTimer != 0) {
            if (AttemptToFinishInquiry()) {
                InquiryCommandTimer = 0;
            } else {
                --InquiryCommandTimer;
                if (InquiryCommandTimer == 0) {
                    GotKeyBoardData(0x7B);
                }
            }
        }
    }
    // Эмуляция клавиатуры конец

    // Работа с остальным
//    private boolean RequestMacOff = false;
//    private boolean RequestMacInterrupt = false;
//    private boolean RequestMacReset = false;
    //

    // Работа с файловой системой и образами начало
    private int MountPending = 0;
    private int disk_op_count;
    
    private final DiskDriver[] Drives = new DiskDriver[NumDrives]; /* open disk image files */

    private void InitDrives() {
        for (int i = 0; i < NumDrives; ++i) {
            Drives[i] = new DiskDriver();
        }
    }

    private int vSonyRead(byte[] Buffer, int offs, int Drive_No, int Sony_Start, int Sony_Count) {
	int result;
        if (Drive_No < NumDrives) {
            if (Drives[Drive_No].isInsert()) {
                result = 0;
                /*System.out.println("Read:"
                                + "\n    - offset:  " + Integer.toHexString(offs)
                                + "\n    - filePos: " + Integer.toHexString(Sony_Start)
                                + "\n    - count:   " + Integer.toHexString(Sony_Count) + "\n");*/
                try {
                    disk_op_count = Drives[Drive_No].read(Buffer, offs, Sony_Start, Sony_Count);
                } catch (IOException ex) {
                    result = -1;
                }
            } else {
                result = 0xFFBF; // Say it's offline (-65)
            }
        } else {
            result = 0xFFC8; // No Such Drive (-56)
        }
        return result;
    }

    private int vSonyWrite(byte[] Buffer, int offs, int Drive_No, int Sony_Start, int Sony_Count) {
        int result;
        if (Drive_No < NumDrives) {
            if (Drives[Drive_No].isInsert()) {
                result = 0;
                try {
                    disk_op_count = Drives[Drive_No].read(Buffer, offs, Sony_Start, Sony_Count);
                } catch (IOException ex) {
                    result = -1;
                }
            } else {
                result = 0xFFBF; // Say it's offline (-65)
            }
        } else {
            result = 0xFFC8; // No Such Drive (-56)
        }
        return result;
    }

    private boolean vSonyDiskLocked(int Drive_No) {
        return false;
    }

    private int vSonyGetSize(int Drive_No) {
        int result;
        if (Drive_No < NumDrives) {
            if (Drives[Drive_No].isInsert()) {
                result = 0;
                try {
                    disk_op_count = Drives[Drive_No].getSize();
                } catch (IOException ex) {
                    result = -1;
                }
                //System.out.println("Get Size: " + Integer.toHexString(disk_op_count) + "\n");
            } else {
                result = 0xFFBF; // Say it's offline (-65)
            }
        } else {
            result = 0xFFC8; // No Such Drive (-56)
        }
        return result;
    }

    private int vSonyEject(int Drive_No) {
        int result;
        if (Drive_No < NumDrives) {
            if (Drives[Drive_No].isInsert()) {
                try {
                    Drives[Drive_No].close();
                } catch (IOException ex) {
                }
            }
            result = 0x0000;
        } else {
            result = 0xFFC8; // No Such Drive (-56)
        }
        return result;
    }

    private int vSonyVerify(int Drive_No) {
        int result;
        //System.out.println("Disk check\n");
        if (Drive_No < NumDrives) {
            if (Drives[Drive_No].isInsert()) {
                result = 0x0000; // No Error (0)
            } else {
                result = 0xFFBF; // Say it's offline (-65)
            }
        } else {
            result = 0xFFC8; // No Such Drive (-56)
        }
        return result;
    }

    private int vSonyFormat(int Drive_No) {
        int result;
        if (Drive_No < NumDrives) {
            if (Drives[Drive_No].isInsert()) {
                result = 0xFFD4; // Write Protected (-44)
            } else {
                result = 0xFFBF; // Say it's offline (-65)
            }
        } else {
            result = 0xFFC8; // No Such Drive (-56)
        }
        return result;
    }

    /*private boolean vSonyInserted(int Drive_No) {
        if (Drive_No >= NumDrives) {
            return false;
        } else {
            return Drives[Drive_No].isInsert();
        }
    }*/

    private int FirstFreeDisk() {
        for (int i = 0; i < NumDrives; i ++) {
            if (!Drives[i].isInsert()) {
                return i;
            }
        }
        return -1;
    }

    /*private boolean AnyDiskInserted() {
        //System.out.println("Any disk inserted check\n");
        for (int i = 0; i < NumDrives; ++i) {
            if (Drives[i].isInsert()) {
                return true;
            }
        }
        return false;
    }*/

    private boolean Sony_Insert0(File file) {
        int Drive_No = FirstFreeDisk();
        if (Drive_No < 0) return false;
        try {
            Drives[Drive_No].insert(file);
            MountPending |= (1 << Drive_No);
        } catch (IOException ex) {
            return false;
        }
        return true;
    }
    // Работа с файловой системой и образами конец

    // SONY floppy disk EMulated DeVice start
    private final static int NumDrives = 3;

    private final byte[] kDSK_Var = new byte[24];
    private final int[] ImageOffset = new int[NumDrives];
    
    private final byte[] Temp = new byte[84];

    private int DelayUntilNextInsert;

    private void Sony_Update() {
        if (DelayUntilNextInsert > 0) {
            DelayUntilNextInsert --;
	} else {
            if (MountPending != 0) {
                int CallBack = do_get_mem_long(kDSK_Var, 20);
                //if (CallBack != 0) {
                    int data;
                    int result;
                    for (int i = 0; i < NumDrives; ++i) {
                        if ((MountPending & (1 << i)) != 0) {
                            MountPending &= ~(1 << i);
                            //Sony_Count = 84;
                            result = vSonyRead(Temp, 0, i, 0, 84);
                            if (result == 0) {
                                if ( /* (Temp[81] == 34 or is that 2) && */(Temp[82] == 1) && (Temp[83] == 0)
                                        && (Temp[0] < 63) && (Temp[63] == 0)) {
                                    /* have old style disk image header */
                                    ImageOffset[i] = 84;
                                } else {
                                    ImageOffset[i] = 0;
                                }

                                data = i;
                                if (vSonyDiskLocked(i)) {
                                    data |= 0x00FF0000;
                                }
                                DelayUntilNextInsert = 20;
                                if (CallBack != 0) DiskInsertedPsuedoException(CallBack, data);
                                return;
                                /* only one disk at a time */
                            }
                        }
                    }
                //}
            }
        }
    }

    private void Sony_Access(int addr) {
        //System.out.println("Disk set var[" + addr + "] = " + Integer.toHexString(DataBus));
        if (!ByteSizeAccess) {
            if ((addr & 1) == 0) {
                if (addr < 24) {
                    if (WriteMemAccess) {
                        System.err.println("Disk set var[" + addr + "] = " + Integer.toHexString(DataBus));
                        do_put_mem_word(kDSK_Var, addr, DataBus);
                        if (addr == 0) { // kDSK_Command
                            /* this is a command */
                            int result;
                            int Drive_No = do_get_mem_word(kDSK_Var, 4);
                            switch (DataBus) {
                                case 0: { // kDSKCmdRead
                                    int Sony_Start = do_get_mem_long(kDSK_Var, 8);
                                    int Sony_Count = do_get_mem_long(kDSK_Var, 12);
                                    int Buffera = do_get_mem_long(kDSK_Var, 16);
                                    result = vSonyRead(ram, Buffera, Drive_No, ImageOffset[Drive_No] + Sony_Start, Sony_Count);
                                    do_put_mem_long(kDSK_Var, 12, disk_op_count);
                                }
                                break;
                                case 1: { // kDSKCmdWrite
                                    int Sony_Start = do_get_mem_long(kDSK_Var, 8);
                                    int Sony_Count = do_get_mem_long(kDSK_Var, 12);
                                    int Buffera = do_get_mem_long(kDSK_Var, 16);
                                    result = vSonyWrite(ram, Buffera, Drive_No, ImageOffset[Drive_No] + Sony_Start, Sony_Count);
                                    do_put_mem_long(kDSK_Var, 12, disk_op_count);
                                }
                                break;
                                case 2: // kDSKCmdEject
                                    result = vSonyEject(Drive_No);
//                                    if (do_get_mem_word(kDSK_Var, 6) != 0) {
//                                        if (!AnyDiskInserted()) {
//                                            RequestMacOff = true;
//                                        }
//                                    }
                                    break;
                                case 3: // kDSKCmdVerify
                                    result = vSonyVerify(Drive_No);
                                    break;
                                case 4: // kDSKCmdFormat
                                    result = vSonyFormat(Drive_No);
                                    break;
                                case 5: { // kDSKCmdGetSize
                                    result = vSonyGetSize(Drive_No);
                                    do_put_mem_long(kDSK_Var, 12, disk_op_count - ImageOffset[Drive_No]);
                                }
                                break;
                                default:
                                    /* unimplemented command */
                                    result = -1;
                                    break;
                            }
                            do_put_mem_word(kDSK_Var, 2, result);
                        }
                    } else {
                        DataBus = do_get_mem_word(kDSK_Var, addr);
                        System.out.println("Disk get var[" + addr + "] = " + Integer.toHexString(DataBus));
                    }
                }
            }
        }
    }

    private void Sony_Reset() {
        for (int i = 23; i >= 0; i--) {
            kDSK_Var[i] = 0;
        }
        DelayUntilNextInsert = 0;
    }
    // SONY floppy disk EMulated DeVice end

    // Versatile Interface Adapter EMulated DEVice (VIA) start

// VIA_Get_ORA : VIA Get Port A Data
// This function queries VIA Port A interfaced hardware about their status
    private int VIA_Get_ORA(int Selection) {
        int Value = 0;
        if ((Selection & 0x80) == 0) { // SCC Wait/Request
            Value |= (VIA_GORA7() << 7);
        }

//        if ((Selection & 0x40) == 0) { // Main/Alternate Screen Buffer
//            Value |= (VIA_GORA6() << 6);
//        }

//        if ((Selection & 0x20) == 0) { // Floppy Disk Line SEL
//            Value |= (VIA_GORA5() << 5);
//        }

//        if ((Selection & 0x10) == 0) { // Overlay/Normal Memory Mapping
//            Value |= (VIA_GORA4() << 4);
//        }

//        if ((Selection & 0x08) == 0) { // Main/Alternate Sound Buffer
//            Value |= (VIA_GORA3() << 3);
//        }

//        if ((Selection & 0x04) == 0) { // Sound Volume Bit 2
//            Value |= (VIA_GORA2() << 2);
//        }

//        if ((Selection & 0x02) == 0) { // Sound Volume Bit 1
//            Value |= (VIA_GORA1() << 1);
//        }

//        if ((Selection & 0x01) == 0) { // Sound Volume Bit 0
//            Value |= VIA_GORA0();
//        }
        return Value;
    }

// VIA_Get_ORB : VIA Get Port B Data
// This function queries VIA Port B interfaced hardware about their status
    private int VIA_Get_ORB(int Selection) {
        int Value = 0;
//        if ((Selection & 0x80) == 0) { // Sound Enable
//            Value |= (VIA_GORB7() << 7);
//        }

//        if ((Selection & 0x40) == 0) { // Video Beam in Display
//            Value |= (VIA_GORB6() << 6);
//        }

//        if ((Selection & 0x20) == 0) { // Mouse Y2
//            Value |= (VIA_GORB5() << 5);
//        }

//        if ((Selection & 0x10) == 0) { // Mouse X2
//            Value |= (VIA_GORB4() << 4);
//        }

        if ((Selection & 0x08) == 0) { // Mouse Button
            Value |= (VIA_GORB3() << 3);
        }

//        if ((Selection & 0x04) == 0) { // RTC Enable
//            Value |= (VIA_GORB2() << 2);
//        }

//        if ((Selection & 0x02) == 0) { // RTC Data Clock
//            Value |= (VIA_GORB1() << 1);
//        }

        if ((Selection & 0x01) == 0) { // RTC Data
            Value |= VIA_GORB0();
        }

        return Value;
    }

    private void VIA_Put_ORA(int Selection, int Data) {
//        if ((Selection & 0x80) != 0) { // SCC Wait/Request
//            VIA_PORA7((Data & 0x80) >> 7);
//        }

        if ((Selection & 0x40) != 0) { // Main/Alternate Screen Buffer
            VIA_PORA6((Data & 0x40) >> 6);
        }

//        if ((Selection & 0x20) != 0) { // Floppy Disk Line SEL
//            VIA_PORA5((Data & 0x20) >> 5);
//        }

        if ((Selection & 0x10) != 0) { // Overlay/Normal Memory Mapping
            VIA_PORA4((Data & 0x10) >> 4);
        }

//        if ((Selection & 0x08) != 0) { // Main/Alternate Sound Buffer
//            VIA_PORA3((Data & 0x08) >> 3);
//        }

//        if ((Selection & 0x04) != 0) { // Sound Volume Bit 2
//            VIA_PORA2((Data & 0x04) >> 2);
//        }

//        if ((Selection & 0x02) != 0) { // Sound Volume Bit 1
//            VIA_PORA1((Data & 0x02) >> 1);
//        }

//        if ((Selection & 0x01) != 0) { // Sound Volume Bit 0
//            VIA_PORA0(Data & 0x01);
//        }
    }

    private void VIA_Put_ORB(int Selection, int Data) {
//        if ((Selection & 0x80) != 0) { // Sound Enable
//            VIA_PORB7((Data & 0x80) >> 7);
//        }

//        if ((Selection & 0x40) != 0) { // Video Beam in Display
//            VIA_PORB6((Data & 0x40) >> 6);
//        }

//        if ((Selection & 0x20) != 0) { // Mouse Y2
//            VIA_PORB5((Data & 0x20) >> 5);
//        }

//        if ((Selection & 0x10) != 0) { // Mouse X2
//            VIA_PORB4((Data & 0x10) >> 4);
//        }

//        if ((Selection & 0x08) != 0) { // Mouse Button
//            VIA_PORB3((Data & 0x08) >> 3);
//        }

        if ((Selection & 0x04) != 0) { // RTC Enable
            VIA_PORB2((Data & 0x04) >> 2);
        }

//        if ((Selection & 0x02) != 0) { // RTC Data Clock
//            VIA_PORB1((Data & 0x02) >> 1);
//        }

        if ((Selection & 0x01) != 0) { // RTC Data
            VIA_PORB0(Data & 0x01);
        }
    }

    private int VIA_ORB;    // Buffer B
    //private int ORA_H;     Buffer A with Handshake
    private int VIA_DDR_B;  // Data Direction Register B
    private int VIA_DDR_A;  // Data Direction Register A
    private int VIA_T1C_L;  // Timer 1 Counter Low
    private int VIA_T1C_H;  // Timer 1 Counter High
    private int VIA_T1L_L;  // Timer 1 Latch Low
    private int VIA_T1L_H;  // Timer 1 Latch High
    private int VIA_T2_L;   // Timer 2 Low
    private int VIA_T2_H;   // Timer 2 High
    private int VIA_SR;     // Shift Register
    private int VIA_ACR;    // Auxiliary Control Register
    private int VIA_PCR;    // Peripheral Control Register
    private int VIA_IFR;    // Interrupt Flag Register
    private int VIA_IER;    // Interrupt Enable Register
    private int VIA_ORA;    // Buffer A

    private int T1_Active;
    private int T2_Active;
    private int SR_Active;

    private boolean VIAInterruptRequest;

    private void VIA_Reset() {
        VIA_ORA = 0x79;
        VIA_DDR_A = 0x7F;
        VIA_ORB = 0x87;
        VIA_DDR_B = 0x87;
        VIA_T1C_L = VIA_T1C_H = VIA_T1L_L = VIA_T1L_H = 0x00;
        VIA_T2_L = VIA_T2_H = VIA_SR = VIA_ACR = 0x00;
        VIA_PCR = VIA_IFR = VIA_IER = 0x00;
        T1_Active = T2_Active = SR_Active = 0x00;
        VIAInterruptRequest = false;
    }

    private void CheckVIAInterruptFlag() {
        boolean NewVIAInterruptRequest = (VIA_IFR & VIA_IER) != 0;
        if (NewVIAInterruptRequest != VIAInterruptRequest) {
            VIAInterruptRequest = NewVIAInterruptRequest;
            if (VIAInterruptRequest) {
                ViaException();
            }
        }
    }

    private void SetVIAInterruptFLag(int VIA_Int) {
        VIA_IFR |= 1 << VIA_Int;
        CheckVIAInterruptFlag();
    }

    private void GotKeyBoardData(int v) {
        VIA_SR = v;
        SetVIAInterruptFLag(2);
    }

    private void VIA_Access(int addr) {
        if (ByteSizeAccess) {
            if ((addr & 1) == 0) {
                switch (addr >> 9) {
                    case 0x00:
                        if ((VIA_PCR & 0xE0) == 0) {
                            VIA_IFR &= 0xF7;
                            CheckVIAInterruptFlag();
                        }
                        if (WriteMemAccess) {
                            VIA_Put_ORB(VIA_DDR_B, DataBus);
                            VIA_ORB = DataBus;
                        } else {
                            VIA_ORB = (VIA_ORB & VIA_DDR_B) + VIA_Get_ORB(VIA_DDR_B);
                            DataBus = VIA_ORB;
                        }
                        break;
                    case 0x02:
                        if (WriteMemAccess) {
                            VIA_DDR_B = DataBus;
                        } else {
                            DataBus = VIA_DDR_B;
                        }
                        break;
                    case 0x03:
                        if (WriteMemAccess) {
                            VIA_DDR_A = DataBus;
                        } else {
                            DataBus = VIA_DDR_A;
                        }
                        break;
                    case 0x04:
                        VIA_IFR &= 0xBF; // Clear T1 Interrupt
                        CheckVIAInterruptFlag();
                        if (WriteMemAccess) {
                            VIA_T1C_L = DataBus;
                        } else {
                            DataBus = VIA_T1C_L;
                        }
                        break;
                    case 0x05:
                        if (WriteMemAccess) {
                            VIA_T1C_H = DataBus;
                        } else {
                            DataBus = VIA_T1C_H;
                        }
                        break;
                    case 0x06:
                        if (WriteMemAccess) {
                            VIA_T1L_L = DataBus;
                        } else {
                            DataBus = VIA_T1L_L;
                        }
                        break;
                    case 0x07:
                        if (WriteMemAccess) {
                            VIA_T1L_H = DataBus;
                            VIA_IFR &= 0xBF; // Clear T1 Interrupt
                            CheckVIAInterruptFlag();
                            VIA_T1C_H = VIA_T1L_H;
                            VIA_T1C_L = VIA_T1L_L;
                            T1_Active = 1;
                        } else {
                            DataBus = VIA_T1L_H;
                        }
                        break;
                    case 0x08:
                        if (WriteMemAccess) {
                            VIA_T2_L = DataBus;
                        } else {
                            VIA_IFR &= 0xDF; // Clear T2 Interrupt
                            CheckVIAInterruptFlag();
                            DataBus = VIA_T2_L;
                        }
                        break;
                    case 0x09:
                        if (WriteMemAccess) {
                            VIA_T2_H = DataBus;
                            VIA_IFR &= 0xDF; // Clear T2 Interrupt
                            CheckVIAInterruptFlag();
                            T2_Active = 1;
                        } else {
                            DataBus = VIA_T2_H;
                        }
                        break;
                    case 0x0A:
                        /* if write access and not SR_Active,
						then VIA.SR won't be changed. is this correct?
                         */
                        if (SR_Active == 1) {
                            switch ((VIA_ACR & 0x1C) >> 2) {
                                case 1:
                                case 2:
                                case 3: // Shifting In
                                    Keyboard_Get();
                                    break;
                                case 4:
                                case 5:
                                case 6:
                                case 7: // Shifting Out
                                    if (WriteMemAccess) {
                                        VIA_SR = DataBus;
                                    }
                                    Keyboard_Put(VIA_SR);
                                    SetVIAInterruptFLag(2);
                                    /* don't wait */
                                    break;
                            }
                            SR_Active = 0;
                        }
                        if (!WriteMemAccess) {
                            DataBus = VIA_SR;
                        }
                        break;
                    case 0x0B:
                        if (WriteMemAccess) {
                            VIA_ACR = DataBus;
                            switch ((VIA_ACR & 0x1C) >> 2) {
                                case 0:
                                    SR_Active = 0;
                                    break;
                                default:
                                    SR_Active = 1;
                                    break;
                            }
                        } else {
                            DataBus = VIA_ACR;
                        }
                        break;
                    case 0x0C:
                        if (WriteMemAccess) {
                            VIA_PCR = DataBus;
                        } else {
                            DataBus = VIA_PCR;
                        }
                        break;
                    case 0x0D:
                        if (WriteMemAccess) {
                            if ((DataBus & 0x80) == 0) {
                                VIA_IFR = VIA_IFR & (0x7F - DataBus); // Clear Enable Bits
                            } else {
                                VIA_IFR = VIA_IFR | (DataBus & 0x7F); // Set Enable Bits
                            }
                            CheckVIAInterruptFlag();
                        } else {
                            DataBus = VIA_IFR;
                            if (VIA_IFR != 0) {
                                DataBus |= 0x80;
                            }
                        }
                        break;
                    case 0x0E:
                        if (WriteMemAccess) {
                            if ((DataBus & 0x80) == 0) {
                                VIA_IER = VIA_IER & (0x7F - DataBus); // Clear Enable Bits
                            } else {
                                VIA_IER = VIA_IER | (DataBus & 0x7F); // Set Enable Bits
                            }
                            CheckVIAInterruptFlag();
                        } else {
                            DataBus = VIA_IER;
                        }
                        break;
                    case 0x0F:
                    case 0x01:
                        if ((VIA_PCR & 0xE) == 0) {
                            VIA_IFR &= 0xFE;
                            CheckVIAInterruptFlag();
                        }
                        if (WriteMemAccess) {
                            VIA_Put_ORA(VIA_DDR_A, DataBus);
                            VIA_ORA = DataBus;
                        } else {
                            VIA_ORA = (VIA_ORA & VIA_DDR_A) + VIA_Get_ORA(VIA_DDR_A);
                            DataBus = VIA_ORA;
                        }
                        break;
                }
            }
        }
    }

    private void VIA_Timer() {
        int Temp;
        if (T1_Active == 1) { // Timer 1 is Active
            Temp = (VIA_T1C_H << 8) + VIA_T1C_L; // Get Timer 1 Counter
            Temp -= 50000;
            if (Temp <= 0) { // Complete?
                if ((VIA_ACR & 0x40) != 0) { // Free Running?
                    VIA_T1C_H = VIA_T1L_H; // Reload Counter from Latches
                    VIA_T1C_L = VIA_T1L_L;
                } else {
                    VIA_T1C_H = VIA_T1C_L = 0;
                    T1_Active = 0; // Deactivate Timer 1
                }
                SetVIAInterruptFLag(6);
            } else {
                VIA_T1C_H = (Temp & 0xFF00) >> 8; // Save Counter
                VIA_T1C_L = Temp & 0xFF;
            }
        }
        if (T2_Active == 1) {
            Temp = (VIA_T2_H << 8) + VIA_T2_L; // Get Timer 1 Counter
            Temp -= 50000;
            if (Temp <= 0) {
                T2_Active = 0; // Deactivate Timer 2
                VIA_T2_H = VIA_T2_L = 0;
                SetVIAInterruptFLag(5);
            } else {
                VIA_T2_H = (Temp & 0xFF00) >> 8; // Save Counter
                VIA_T2_L = Temp & 0xFF;
            }
            
        }
    }

    private void VIA_Int_Vertical_Blanking() {
        SetVIAInterruptFLag(1);
    }

    private void VIA_Int_One_Second() {
        SetVIAInterruptFLag(0);
    }
    // Versatile Interface Adapter EMulated DEVice (VIA) end

    // Real Time Clock EMulated DEVice (RTC) start
    // RTC VIA Flags
    private int RTC_Enabled;
    private int RTC_Clock;

    // RTC Data
    private int RTC_Mode;
    private int RTC_Command;
    private int RTC_Data;
    private int RTC_Counter;

    // RTC Registers
    private final int RTC_Seconds_1[] = new int[4]; // Apple Documented Seconds
    private final int RTC_Seconds_2[] = new int[4]; // Unknown (apparent) second Seconds backup
    private final int RTC_PARAMRAM[] = new int[20];

    private void RTC_Init() {
        int Counter;
        int secs;
        RTC_Enabled = RTC_Clock = RTC_Mode = RTC_Command = RTC_Data = RTC_Counter = 0;
        secs = (int) (System.currentTimeMillis() / 1000L);
        RTC_Seconds_1[0] = secs & 0xFF;
        RTC_Seconds_1[1] = (secs & 0xFF00) >> 8;
        RTC_Seconds_1[2] = (secs & 0xFF0000) >> 16;
        RTC_Seconds_1[3] = secs >>> 24;

        RTC_Seconds_2[0] = RTC_Seconds_1[0];
        RTC_Seconds_2[1] = RTC_Seconds_1[1];
        RTC_Seconds_2[2] = RTC_Seconds_1[2];
        RTC_Seconds_2[3] = RTC_Seconds_1[3];

        for (Counter = 0; Counter < 20; Counter++) {
            RTC_PARAMRAM[Counter] = 0;
        }
        RTC_PARAMRAM[0] = 168;
        RTC_PARAMRAM[3] = 34;
        RTC_PARAMRAM[4] = 204;
        RTC_PARAMRAM[5] = 10;
        RTC_PARAMRAM[6] = 204;
        RTC_PARAMRAM[7] = 10;
        RTC_PARAMRAM[13] = 2;
        RTC_PARAMRAM[14] = 99;
        RTC_PARAMRAM[17] = 83;
        RTC_PARAMRAM[18] = 4;
        RTC_PARAMRAM[19] = 76;
    }

    private void RTC_Interrupt() {
        int Seconds;
        Seconds = (RTC_Seconds_1[3] << 24) + (RTC_Seconds_1[2] << 16) + (RTC_Seconds_1[1] << 8) + RTC_Seconds_1[0];
        Seconds++;
        RTC_Seconds_1[0] = Seconds & 0xFF;
        RTC_Seconds_1[1] = (Seconds & 0xFF00) >> 8;
        RTC_Seconds_1[2] = (Seconds & 0xFF0000) >> 16;
        RTC_Seconds_1[3] = Seconds >>> 24;

        Seconds = (RTC_Seconds_2[3] << 24) + (RTC_Seconds_2[2] << 16) + (RTC_Seconds_2[1] << 8) + RTC_Seconds_2[0];
        Seconds++;
        RTC_Seconds_2[0] = Seconds & 0xFF;
        RTC_Seconds_2[1] = (Seconds & 0xFF00) >> 8;
        RTC_Seconds_2[2] = (Seconds & 0xFF0000) >> 16;
        RTC_Seconds_2[3] = Seconds >>> 24;
    }

    private void RTC_Put(int in) {
        RTC_Data = (RTC_Data << 1) + in;
        RTC_Counter++;
        if (RTC_Counter == 8) {
            RTC_DoCmd();
        }
    }

    private int RTC_Get() {
        return ((RTC_Data >> --RTC_Counter) & 0x01);
    }

    private void RTC_DoCmd() {
        if (RTC_Mode == 0) { // This Byte is a RTC Command
            RTC_Command = RTC_Data;
            RTC_Data = 0;
            if ((RTC_Command & 0x80) == 0x00) { // Write Command
                RTC_Mode = 1; // Set to switch to Write Mode
                RTC_Counter = 0; // Reset the Counter
            } else { // Read Command
                RTC_Read_Reg(); // Read RTC Register
            }
        } else { // Data has come in that needs to be written
            RTC_Write_Reg();
        }
    }

    private void RTC_Read_Reg() {
        switch ((RTC_Command & 0x7C) >> 2) {
            case 0:
                RTC_Data = RTC_Seconds_1[0];
                break;
            case 1:
                RTC_Data = RTC_Seconds_1[1];
                break;
            case 2:
                RTC_Data = RTC_Seconds_1[2];
                break;
            case 3:
                RTC_Data = RTC_Seconds_1[3];
                break;
            case 4:
                RTC_Data = RTC_Seconds_2[0];
                break;
            case 5:
                RTC_Data = RTC_Seconds_2[1];
                break;
            case 6:
                RTC_Data = RTC_Seconds_2[2];
                break;
            case 7:
                RTC_Data = RTC_Seconds_2[3];
                break;
            case 8:
            case 9:
            case 10:
            case 11:
                RTC_Data = RTC_PARAMRAM[((RTC_Command & 0x0C) >> 2) + 0x10];
                break;
            case 16:
            case 17:
            case 18:
            case 19:
            case 20:
            case 21:
            case 22:
            case 23:
            case 24:
            case 25:
            case 26:
            case 27:
            case 28:
            case 29:
            case 30:
            case 31:
                RTC_Data = RTC_PARAMRAM[((RTC_Command & 0x3C) >> 2)];
                break;
            default:
                RTC_Data = 0;
                /* Debugger(); */
                break;
        }
        RTC_Command = RTC_Mode = 0;
    }

    private void RTC_Write_Reg() {
        switch ((RTC_Command & 0x7C) >> 2) {
            case 0:
                RTC_Seconds_1[0] = RTC_Data;
                break;
            case 1:
                RTC_Seconds_1[1] = RTC_Data;
                break;
            case 2:
                RTC_Seconds_1[2] = RTC_Data;
                break;
            case 3:
                RTC_Seconds_1[3] = RTC_Data;
                break;
            case 4:
                RTC_Seconds_2[0] = RTC_Data;
                break;
            case 5:
                RTC_Seconds_2[1] = RTC_Data;
                break;
            case 6:
                RTC_Seconds_2[2] = RTC_Data;
                break;
            case 7:
                RTC_Seconds_2[3] = RTC_Data;
                break;
            case 8:
            case 9:
            case 10:
            case 11:
                RTC_PARAMRAM[((RTC_Command & 0x0C) >> 2) + 0x10] = RTC_Data;
                break;
            case 12:
                break; // Test Write, do nothing
            case 13:
                /* RTC_Save(); */
                break; // Write_Protect Register
            case 16:
            case 17:
            case 18:
            case 19:
            case 20:
            case 21:
            case 22:
            case 23:
            case 24:
            case 25:
            case 26:
            case 27:
            case 28:
            case 29:
            case 30:
            case 31:
                RTC_PARAMRAM[((RTC_Command & 0x3C) >> 2)] = RTC_Data;
                break;
            default:
                RTC_Data = 0;
                break;
        }
        RTC_Command = RTC_Data = RTC_Counter = RTC_Mode = 0;
    }

    private void VIA_PORB2(int Data) {
        RTC_Enabled = Data;
    }

    private int VIA_GORB0() { // RTC Data
        if ((RTC_Enabled == 0) && (RTC_Clock == 1)) {
            return RTC_Get();
        } else {
            return 0;
        }
    }

    private void VIA_PORB0(int Data) {
        if ((RTC_Enabled == 0) && (RTC_Clock == 1)) {
            RTC_Put(Data);
        }
    }
    // Real Time Clock EMulated DEVice (RTC) end

    // Serial Communications Controller EMulated DEVice (SCC) start
    private final int[] SCC_A_RR = new int[16];
    private final int[] SCC_A_WR = new int[16];
    private final int[] SCC_B_RR = new int[16];
    private final int[] SCC_B_WR = new int[16];
    private int SCC_Mode;
    private int SCC_Reg;

    private boolean Mouse_Enabled() {
        return ((SCC_A_WR[9] & 0x08) != 0x00);
    }

    private void SCC_Reset() {
        int Counter;

        SCC_Mode = SCC_Reg = 0;

        for (Counter = 0; Counter < 16; Counter++) {
            SCC_A_RR[Counter] = 0;
            SCC_B_RR[Counter] = 0;
            SCC_A_WR[Counter] = 0;
            SCC_B_WR[Counter] = 0;
        }
        SCC_A_RR[0] = 0x04;
        SCC_B_RR[0] = 0x04;
    }

    private int SCC_GetAReg() {
        return SCC_A_RR[SCC_Reg];
    }

    private void SCC_PutAReg(int in) {
        SCC_A_WR[SCC_Reg] = in;
    }

    private int SCC_GetBReg() {
        return SCC_B_RR[SCC_Reg];
    }

    private void SCC_PutBReg(int in) {
        SCC_B_WR[SCC_Reg] = in;
    }

    private void SCC_Access(int addr) {
        if (ByteSizeAccess) {
            if ((addr & 1) == 0) {
                switch (addr >> 1) {
                    case 0: // Channel B Control
                        if (SCC_Mode == 0) {
                            if (WriteMemAccess) {
                                SCC_Mode = 1;
                                SCC_Reg = DataBus & 0x0F;
                                SCC_PutBReg(DataBus & 0xF0);
                            } else {
                                SCC_Reg = 0;
                                DataBus = SCC_GetBReg();
                            }
                        } else {
                            SCC_Mode = 0;
                            if (WriteMemAccess) {
                                SCC_PutBReg(DataBus);
                            } else {
                                DataBus = SCC_GetBReg();
                            }
                        }
                        break;
                    case 1: // Channel A Control
                        if (SCC_Mode == 0) {
                            if (WriteMemAccess) {
                                SCC_Mode = 1;
                                SCC_Reg = DataBus & 0x0F;
                                SCC_PutAReg(DataBus & 0xF0);
                            } else {
                                SCC_Reg = 0;
                                DataBus = SCC_GetAReg();
                            }
                        } else {
                            SCC_Mode = 0;
                            if (WriteMemAccess) {
                                SCC_PutAReg(DataBus);
                            } else {
                                DataBus = SCC_GetAReg();
                            }
                        }
                        break;
                    case 2: // Channel B Data
                        break;
                    case 3: // Channel A Data
                        break;
                }
            } else if (!WriteMemAccess) {
                SCC_Reset();
            }
        }
    }

    // VIA Interface Functions
    private int VIA_GORA7() {
        return 1; // No Wait/Requests
    }
    // Serial Communications Controller EMulated DEVice (SCC) end

    // Small Computer System Interface EMulated DeVice (SCSI) start
    private final int[] SCSI = new int[0x00010];
    
    private void SCSI_Reset() {
        for (int i = 0xF; i >= 0; i--) {
            SCSI[i] = 0;
        }
    }

    private void SCSI_BusReset() {
        SCSI[0 + 0] = 0;
        SCSI[1 + 0] = 0;
        SCSI[0 + 2] = 0x80;
        SCSI[1 + 2] &= 0x80;
        SCSI[0 + 4] &= 0x40;
        SCSI[1 + 4] &= 0x40;
        SCSI[0 + 6] = 0;
        SCSI[1 + 6] = 0;
        SCSI[0 + 8] = 0x80;
        SCSI[1 + 8] = 0;
        SCSI[0 + 10] = 0x10;
        SCSI[1 + 10] = 0;
        SCSI[0 + 12] = 0;
        SCSI[1 + 12] = 0;
        SCSI[0 + 14] = 0;
        SCSI[1 + 14] = 0;
        put_word(0xb22, get_word(0xb22) | 0x8000
    );
}

    private void SCSI_Check() {
        // The arbitration select/reselect scenario [stub.. doesn't really work...]
        if ((SCSI[1 + 0] >> 7) == 1) {  // Check if the Mac tries to be an initiator
            if ((SCSI[1 + 4] & 1) == 1) { // the Mac set arbitration in progress
                // stub! tell the mac that there is arbitration in progress...
                SCSI[0 + 2] |= 0x40;
                // ... that we didn't lose arbitration ...
                SCSI[0 + 2] &= ~0x20;
                // ... and that there isn't a higher priority ID present...
                SCSI[0 + 0] = 0x00;

                // ... the arbitration and selection/reselection is complete.
                // the initiator tries to connect to the SCSI device, fails
                // and returns after timeout.
            }
        }

        // check the chip registers, AS SET BY THE CPU
        if ((SCSI[1 + 2] >> 7) == 1) {  // Check Assert RST
            SCSI_BusReset();
        } else {
            SCSI[0 + 2] &= ~0x80;
            SCSI[0 + 8] &= ~0x80;
        }

        if ((SCSI[1 + 2] >> 2) == 1) { // Check Assert SEL
            SCSI[0 + 8] |= 0x02;
            SCSI[0 + 10] = 0x10;
        } else {
            SCSI[0 + 8] &= ~0x02;
        }
    }

    private void SCSI_Access(int addr) {
        if ((addr & 0xE) == 0) {
            addr = (addr >> 3) | (addr & 1);
            if (addr < 0x00010) {
                if (ByteSizeAccess) {
                    if (WriteMemAccess) {
                        SCSI[addr] = DataBus;
                        SCSI_Check();
                    } else {
                        DataBus = SCSI[addr];
                    }
                }
            }
        }
    }
    // Small Computer System Interface EMulated DeVice (SCSI) end

    // Integrated Woz Macine EMulated DEVice (IWM) start
//    private int vSel;
    private int IWM_DataIn;    // Read Data Register
    private int IWM_Handshake; // Read Handshake Register
    private int IWM_Status;    // Read Status Register
    private int IWM_Mode;      // Drive Off : Write Mode Register
    // Drive On  : Write Data Register
//    private int IWM_DataOut;   // Write Data Register
    private int IWM_Lines;     // Used to Access Disk Drive Registers

    private void IWM_Reset() {
        IWM_DataIn = IWM_Handshake = IWM_Status = IWM_Mode = /*IWM_DataOut =*/ IWM_Lines = 0;
    }

    private void IWM_Set_Lines(int line, int the_mode) {
        if (the_mode == 0) {
            IWM_Lines &= (0xFF - line);
        } else {
            IWM_Lines |= line;
        }
    }

    private void IWM_Access(int addr) {
        //System.out.println("IWM_Access\n - addr = 0x" + Integer.toHexString(addr >> 9) +
        //        "\n - data = " + Integer.toHexString(DataBus) +
        //        "\n - wr = " + WriteMemAccess);
        if (ByteSizeAccess) {
            if ((addr & 1) == 0) {
                switch (addr >> 9) {
                    case 0x00:
                        IWM_Set_Lines(0x01, 0);
                        break;
                    case 0x01:
                        IWM_Set_Lines(0x01, 1);
                        break;
                    case 0x02:
                        IWM_Set_Lines(0x02, 0);
                        break;
                    case 0x03:
                        IWM_Set_Lines(0x02, 1);
                        break;
                    case 0x04:
                        IWM_Set_Lines(0x04, 0);
                        break;
                    case 0x05:
                        IWM_Set_Lines(0x04, 1);
                        break;
                    case 0x06:
                        IWM_Set_Lines(0x08, 0);
                        break;
                    case 0x07:
                        IWM_Set_Lines(0x08, 1);
                        break;
                    case 0x08:
                        IWM_Status &= 0xDF;
                        IWM_Set_Lines(0x10, 0);
                        break;
                    case 0x09:
                        IWM_Status |= 0x20;
                        IWM_Set_Lines(0x10, 1);
                        break;
                    case 0x0A:
                        IWM_Set_Lines(0x20, 0);
                        break;
                    case 0x0B:
                        IWM_Set_Lines(0x20, 1);
                        break;
                    case 0x0C:
                        IWM_Set_Lines(0x40, 0);
                        break;
                    case 0x0D:
                        IWM_Set_Lines(0x40, 1);
                        break;
                    case 0x0E:
                        if (!WriteMemAccess) {
                            DataBus = IWM_Read_Reg();
                        }
                        IWM_Set_Lines(0x80, 0);
                        break;
                    case 0x0F:
                        if (WriteMemAccess) {
                            IWM_Write_Reg(DataBus);
                        }
                        IWM_Set_Lines(0x80, 1);
                        break;
                }
            }
        }
    }

    private int IWM_Read_Reg() {
        switch ((IWM_Lines & 0x60) >> 6) {
            case 0:
                return IWM_DataIn;
            case 1:
                return IWM_Status;
            case 2:
                return IWM_Handshake;
            case 3:
            default:
                /* should alway be in 0-3, but compiler warnings don't know that */
                return 0;
        }
    }

    private void IWM_Write_Reg(int in) {
        if (((IWM_Lines & 0x10) >> 4) == 0) {
            IWM_Mode = in;
            IWM_Status = ((IWM_Status & 0xE0) + (IWM_Mode & 0x1F));
        }
    }

//    private void VIA_PORA5(int Data) {
//        vSel = Data;
//    }
    // Integrated Woz Macine EMulated DEVice (IWM) end

    // Random memory access start
    private int AddressBus;
    private int DataBus;
    private boolean ByteSizeAccess;
    private boolean WriteMemAccess;
    //

    private void MM_Access() {
        /*if (AddressBus > 0x00f00000) {
            System.out.println("opcode = " + Integer.toHexString(opcode) + ", pc = 0x" + Integer.toHexString(regs_pc));
            System.out.println("MM_Access addr = 0x" + Integer.toHexString(AddressBus) + ", wr = " + WriteMemAccess);
        }*/
        if (AddressBus < 0x00C00000) {
            if (AddressBus < 0x00800000) {
                if ((AddressBus >= 0x00580000) && (AddressBus < 0x00600000)) {
                    SCSI_Access(AddressBus - 0x00580000);
                }
            } else if (AddressBus >= 0x00A00000) {
                if (WriteMemAccess) {
                    /* if ((AddressBus >= kSCCWr_Block_Base) && (AddressBus < kSCCWr_Block_Top)) */  // Write Only Address
                    {
                        SCC_Access((AddressBus - 0xBFFFF9) & 0x07);
                    }
                }
            } else if (!WriteMemAccess) {
                /* if ((AddressBus >= kSCCRd_Block_Base) && (AddressBus < kSCCRd_Block_Top)) */ // Read Only Address
                {
                    SCC_Access((AddressBus - 0x9FFFF8) & 0x07);
                }
            }
        } else if (AddressBus < 0x00E80000) {
            if (/* (AddressBus >= kIWM_Block_Base) && */(AddressBus < 0x00E00000)) {
                IWM_Access((AddressBus - 0xDFE1FF) & 0x001FFF);
            }
        } else if (/* (AddressBus >= kVIA_Block_Base) && */(AddressBus < 0x00F00000)) {
            VIA_Access((AddressBus - 0xEFE1FE) % 0x2000);
        } else if ((AddressBus >= 0x00F40000) && (AddressBus < 0x00F40020)) {
            Sony_Access(AddressBus - 0x00F40000);
        } else {
            if ((AddressBus >= 0x00FFFFF0)) { // && (AddressBus < 0x01000000)) {
                SetAutoVector();
                /* Exception(regs.intmask+24, 0); */
            }
        }
    }

    private void Memory_Reset() {
        VIA_PORA4(1);
    }

    private void VIA_PORA4(int Data) {
        if (Data != 0) {
            bankR[3] = bankW[3] = ram;
            bank_offs[3] = 0x00600000;
            bankR[0] = bankR[2] = rom;
            bank_offs[2] = 0x00400000;
        } else {
            bankR[0] = bankW[0] = ram;
            if (ram.length >= 0x00200000) {
                bankR[1] = bankW[1] = ram;
            }
        }
    }

    //
    private int do_get_mem_word(byte[] mem, int addr) {
        return (mem[addr + 1] & 0xFF) | ((mem[addr] & 0xFF) << 8);
    }

    private int do_get_mem_long(byte[] mem, int addr) {
        return (mem[addr + 3] & 0xFF) | ((mem[addr + 2] & 0xFF) << 8) |
               ((mem[addr + 1] & 0xFF) << 16) | ((mem[addr] & 0xFF) << 24);
    }

    private void do_put_mem_word(byte[] mem, int addr, int v) {
        mem[addr + 1] = (byte) v;
        mem[addr] = (byte) (v >> 8);
    }

    private void do_put_mem_long(byte[] mem, int addr, int v) {
        mem[addr + 3] = (byte) v;
        mem[addr + 2] = (byte) (v >> 8);
        mem[addr + 1] = (byte) (v >> 16);
        mem[addr] = (byte) (v >> 24);
    }
    //
    private final byte[][] bankR = new byte[4][];
    private final byte[][] bankW = new byte[4][];
    private final int[] bank_offs = new int[4];
    //
    private int get_byte(int addr) {
        addr &= 0x00ffFFFF;
        byte[] bank = null;
        if (addr < 0x00700000) {
            int i = addr >> 21;
            if ((bank = bankR[i]) != null) {
                i = addr - bank_offs[i];
                if (i >= bank.length) bank = null;
                else addr = i;
            }
        }
        if (bank == null) {
            AddressBus = addr;
            DataBus = 0;
            ByteSizeAccess = true;
            WriteMemAccess = false;
            MM_Access();
            return DataBus & 0xFF;
        }
        return bank[addr] & 0xFF;
    }

    private int get_word(int addr) {
        addr &= 0x00ffFFFF;
        byte[] bank = null;
        if (addr < 0x00700000) {
            int i = addr >> 21;
            if ((bank = bankR[i]) != null) {
                i = addr - bank_offs[i];
                if (i >= bank.length) bank = null;
                else addr = i;
            }
        }
        if (bank == null) {
            AddressBus = addr;
            DataBus = 0;
            ByteSizeAccess =
            WriteMemAccess = false;
            MM_Access();
            return DataBus & 0xFFFF;
        }
        return (bank[addr + 1] & 0xFF) | ((bank[addr] & 0xFF) << 8);
    }
    
    private int get_long(int addr) {
        addr &= 0x00ffFFFF;
        byte[] bank = null;
        if (addr < 0x00700000) {
            int i = addr >> 21;
            if ((bank = bankR[i]) != null) {
                i = addr - bank_offs[i];
                if (i >= bank.length) bank = null;
                else addr = i;
            }
        }
        if (bank == null) {
            AddressBus = addr;
            DataBus = 0;
            ByteSizeAccess =
            WriteMemAccess = false;
            MM_Access();
            int hi = DataBus;
            AddressBus = addr + 2;
            DataBus = 0;
            ByteSizeAccess =
            WriteMemAccess = false;
            MM_Access();
            return (hi << 16) | (DataBus & 0xFFFF);
        }
        return (bank[addr + 3] & 0xFF) | ((bank[addr + 2] & 0xFF) << 8) |
               ((bank[addr + 1] & 0xFF) << 16) | ((bank[addr] & 0xFF) << 24);
    }

    private void put_byte(int addr, int v) {
        addr &= 0x00ffFFFF;
        byte[] bank = null;
        if (addr < 0x00700000) {
            int i = addr >> 21;
            if ((bank = bankW[i]) != null) {
                i = addr - bank_offs[i];
                if (i >= bank.length) bank = null;
                else addr = i;
            }
        }
        if (bank != null) {
            bank[addr] = (byte) v;
        } else {
            AddressBus = addr;
            DataBus = v & 0xFF;
            ByteSizeAccess =
            WriteMemAccess = true;
            MM_Access();
        }        
    }

    private void put_word(int addr, int v) {
        addr &= 0x00ffFFFF;
        byte[] bank = null;
        if (addr < 0x00700000) {
            int i = addr >> 21;
            if ((bank = bankW[i]) != null) {
                i = addr - bank_offs[i];
                if (i >= bank.length) bank = null;
                else addr = i;
            }
        }
        if (bank != null) {
            bank[addr + 1] = (byte) v;
            bank[addr] = (byte) (v >> 8);
        } else {
            AddressBus = addr;
            DataBus = v & 0xFFFF;
            ByteSizeAccess = false;
            WriteMemAccess = true;
            MM_Access();
        }
    }
    
    private void put_long(int addr, int v) {
        addr &= 0x00ffFFFF;
        byte[] bank = null;
        if (addr < 0x00700000) {
            int i = addr >> 21;
            if ((bank = bankW[i]) != null) {
                i = addr - bank_offs[i];
                if (i >= bank.length) bank = null;
                else addr = i;
            }
        }
        if (bank != null) {
            bank[addr + 3] = (byte) v;
            bank[addr + 2] = (byte) (v >> 8);
            bank[addr + 1] = (byte) (v >> 16);
            bank[addr] = (byte) (v >> 24);
        } else {
            AddressBus = addr;
            DataBus = v >>> 16;
            ByteSizeAccess = false;
            WriteMemAccess = true;
            MM_Access();
            AddressBus = addr + 2;
            DataBus = v & 0xFFFF;
            ByteSizeAccess = false;
            WriteMemAccess = true;
            MM_Access();
        }
    }
    // Random memory access end

    // Motorola 68K эмулятор начало
    private final int regs[] = new int[16];
    private int regs_usp;
    private int regs_isp;
    private int regs_pc;
    private int regs_intmask;
    private boolean regs_t1;
    private boolean regs_s;
    private boolean regs_m;
    private boolean regs_AutoVectorPending;
    private boolean regs_TracePending;
    private boolean regs_ExternalInterruptPending;

    private boolean ZFLG;
    private boolean NFLG;
    private boolean CFLG;
    private boolean VFLG;
    private boolean XFLG;

    private boolean cctrue( int cc) {
        switch(cc){
            case 0:  return true;                    /* T */
            case 1:  return false;                   /* F */
            case 2:  return !CFLG && !ZFLG;          /* HI */
            case 3:  return CFLG || ZFLG;            /* LS */
            case 4:  return !CFLG;                   /* CC */
            case 5:  return CFLG;                    /* CS */
            case 6:  return !ZFLG;                   /* NE */
            case 7:  return ZFLG;                    /* EQ */
            case 8:  return !VFLG;                   /* VC */
            case 9:  return VFLG;                    /* VS */
            case 10: return !NFLG;                   /* PL */
            case 11: return NFLG;                    /* MI */
            case 12: return NFLG == VFLG;            /* GE */
            case 13: return NFLG != VFLG;            /* LT */
            case 14: return !ZFLG && (NFLG == VFLG); /* GT */
            case 15: return ZFLG || (NFLG != VFLG);  /* LE */
            default: return false; /* shouldn't get here */
        }
    }

    private int m68k_getSR() {
        return (regs_t1 ? 0x8000 : 0)
                    | (regs_s ? 0x2000 : 0) | (regs_m ? 0x1000 : 0) | (regs_intmask << 8)
                    | (XFLG ? 16 : 0) | (NFLG ? 8 : 0) | (ZFLG ? 4 : 0) | (VFLG ? 2 : 0)
                    |  (CFLG ? 1 : 0);
    }
    
    private void m68k_setCR(int newcr) {
        XFLG = (newcr & 16) != 0;
        NFLG = (newcr & 8) != 0;
        ZFLG = (newcr & 4) != 0;
        VFLG = (newcr & 2) != 0;
        CFLG = (newcr & 1) != 0;
    }

    private void m68k_setSR(int newsr) {
        boolean olds = regs_s;
        int oldintmask = regs_intmask;

        m68k_setCR(newsr);
        regs_t1 = (newsr & 0x8000) != 0;
        regs_s = (newsr & 0x2000) != 0;
        regs_m = (newsr & 0x1000) != 0;
        regs_intmask = (newsr >> 8) & 7;
        if (olds != regs_s) {
            if (olds) {
                regs_isp = regs[15];
                regs[15] = regs_usp;
            } else {
                regs_usp = regs[15];
                regs[15] = regs_isp;
            }
        }
        if (regs_intmask != oldintmask) {
            SetExternalInterruptPending();
        }
        if (regs_t1) {
            NeedToGetOut();
        } else {
            regs_TracePending = false;
        }
    }

    private int nextibyte() {
        int r = get_byte(regs_pc + 1);
        regs_pc += 2;
        return r;
    }

    private int nextiword() {
        int r = get_word(regs_pc);
        regs_pc += 2;
        return r;
    }

    private int nextilong() {
        int r = get_long(regs_pc);
        regs_pc += 4;
        return r;
    }

    private void m68k_setpc(int newpc) {
        regs_pc = newpc & 0x00ffFFFF;
    }
    
    private void ExceptionTo(int newpc) {
        int saveSR = m68k_getSR();
        if (!regs_s) {
            regs_usp = regs[15];
            regs[15] = regs_isp;
            regs_s = true;
        }
        regs[15] -= 4;
        put_long(regs[15], regs_pc);
        regs[15] -= 2;
        put_word(regs[15], saveSR);
        m68k_setpc(newpc);
        regs_t1 = regs_m = regs_TracePending = false;
    }

    private void Exception(int nr) {
        ExceptionTo(get_long(nr << 2));
    }
    
    private void DiskInsertedPsuedoException(int newpc, int data) {
        ExceptionTo(newpc);
        regs[15] -= 4;
        put_long(regs[15], data);
    }

    private void DoCheckExternalInterruptPending() {
        int level = VIAInterruptRequest ? 1 : 0;
        if ((level > regs_intmask) || (level == 7)) {
            Exception(24 + level);
            regs_intmask = level;
        }
    }

    private void ViaException() {
        int level = VIAInterruptRequest ? 1 : 0;
        if ((level > regs_intmask) || (level == 7)) {
            SetExternalInterruptPending();
        }
    }

    private void m68k_reset() {
        customreset();
        // Sets the MC68000 reset jump vector...
        m68k_setpc(get_long(0x00000004));
        // Sets the initial stack vector...
        regs[15] = get_long(0x00000000);

        regs_s = true;
        regs_m = regs_t1 =
        ZFLG = CFLG = NFLG = VFLG =
        regs_AutoVectorPending =
        regs_ExternalInterruptPending =
        regs_TracePending = false;
        regs_intmask = 7;
    }

    private void MacInterrupt() {
        Exception(28);
    }
    
    private int get_disp_ea(int base) {
        int dp = nextiword();
        int reg = (dp >> 12) & 15;
        int regd = regs[reg];
        if ((dp & 0x800) == 0) {
            regd = (short) regd;
        }
        return base + ((byte) dp) + regd;
    }

    private static final int AKMemory = 0;
    private static final int AKAddrReg = 1;
    private static final int AKDataReg = 2;
    private static final int AKConstant = 3;
    private static final int AKCCR = 4;
    private static final int AKSR = 5;


    private int opsize, ArgKind, ArgAddr;
    
    private void DecodeModeRegister(int themode, int thereg) {
        switch (themode) {
            case 0:
                ArgKind = AKDataReg;
                ArgAddr = thereg;
                break;
            case 1:
                ArgKind = AKAddrReg;
                ArgAddr = thereg;
                break;
            case 2:
                ArgKind = AKMemory;
                ArgAddr = regs[thereg + 8];
                break;
            case 3:
                ArgKind = AKMemory;
                ArgAddr = regs[thereg + 8];
                if ((thereg == 7) && (opsize == 1)) {
                    regs[thereg + 8] += 2;
                } else {
                    regs[thereg + 8] += opsize;
                }
                break;
            case 4:
                ArgKind = AKMemory;
                if ((thereg == 7) && (opsize == 1)) {
                    regs[thereg + 8] -= 2;
                } else {
                    regs[thereg + 8] -= opsize;
                }
                ArgAddr = regs[thereg + 8];
                break;
            case 5:
                ArgKind = AKMemory;
                ArgAddr = regs[thereg + 8] + (short) nextiword();
                break;
            case 6:
                ArgKind = AKMemory;
                ArgAddr = get_disp_ea(regs[thereg + 8]);
                break;
            case 7:
                switch (thereg) {
                    case 0:
                        ArgKind = AKMemory;
                        ArgAddr = (short) nextiword();
                        break;
                    case 1:
                        ArgKind = AKMemory;
                        ArgAddr = nextilong();
                        break;
                    case 2:
                        ArgKind = AKMemory;
                        ArgAddr = regs_pc;
                        ArgAddr += (short) nextiword();
                        break;
                    case 3:
                        ArgKind = AKMemory;
                        ArgAddr = get_disp_ea(regs_pc);
                        break;
                    case 4:
                        ArgKind = AKConstant;
                        switch (opsize) {
                            case 1:
                                ArgAddr = (byte) nextibyte();
                                break;
                            case 2:
                                ArgAddr = (short) nextiword();
                                break;
                            case 4:
                                ArgAddr = nextilong();
                                break;
                        }
                        break;
                }
                break;
            case 8:
                ArgKind = AKConstant;
                ArgAddr = thereg;
                break;
            case 10:
                ArgKind = AKCCR;
                break;
            case 11:
                ArgKind = AKSR;
                break;
        }
    }

    private int GetArgValue() {
        int v;
        switch (ArgKind) {
            case AKMemory:
                switch (opsize) {
                    case 1:
                        v = (byte) get_byte(ArgAddr);
                        break;
                    case 2:
                        v = (short) get_word(ArgAddr);
                        break;
                    case 4:
                    default:
                        /* for compiler. should be 1, 2, or 4 */
                        v = get_long(ArgAddr);
                        break;
                }
                break;
            case AKAddrReg:
                v = regs[ArgAddr + 8];
                break;
            case AKDataReg:
                v = regs[ArgAddr];
                switch (opsize) {
                    case 1:
                        v = (byte) v;
                        break;
                    case 2:
                        v = (short) v;
                        break;
                    case 4:
                        break;
                }
                break;
            case AKConstant:
                v = ArgAddr;
                break;
            case AKCCR:
                v = m68k_getSR() & 0xFF;
                break;
            case AKSR:
            default:
                v = m68k_getSR();
                break;
        }
        return v;
    }
 
    private void SetArgValue(int v) {
        switch (ArgKind) {
            case AKMemory:
                switch (opsize) {
                    case 1:
                        put_byte(ArgAddr, v);
                        break;
                    case 2:
                        put_word(ArgAddr, v);
                        break;
                    case 4:
                        put_long(ArgAddr, v);
                        break;
                }
                break;
            case AKDataReg:
                switch (opsize) {
                    case 1:
                        regs[ArgAddr] = (regs[ArgAddr] & 0xffffFF00) | ((v) & 0xff);
                        break;
                    case 2:
                        regs[ArgAddr] = (regs[ArgAddr] & 0xffff0000) | ((v) & 0xffff);
                        break;
                    case 4:
                        regs[ArgAddr] = v;
                        break;
                }
                break;
            case AKAddrReg:
                /* stricly speaking, illegal if opsize == 1 */
                regs[ArgAddr + 8] = v;
                break;
            case AKCCR:
                m68k_setCR(v);
                break;
            case AKSR:
                m68k_setSR(v);
                break;
            default:
                //op_illg();
                regs_pc -= 2;
                Exception(4);
                break;
        }
    }
    
    private void DoMove(int m1, int r1, int m2, int r2) /* MOVE */ {
        int src;
        DecodeModeRegister(m1, r1);
        src = GetArgValue();
        DecodeModeRegister(m2, r2);
        if (ArgKind != AKAddrReg) {
            VFLG = CFLG = false;
            ZFLG = (src == 0);
            NFLG = (src < 0);
        }
        SetArgValue(src);
    }

    private void DoMoveNoFlags(int m1, int r1, int m2, int r2) /* MOVE */ {
        int src;
        DecodeModeRegister(m1, r1);
        src = GetArgValue();
        DecodeModeRegister(m2, r2);
        SetArgValue(src);
    }

    private void DoBinOp1(int m1, int r1, int m2, int r2, int binop) {
        int srcvalue;
        int dstvalue;
        DecodeModeRegister(m1, r1);
        srcvalue = GetArgValue();
        DecodeModeRegister(m2, r2);
        dstvalue = GetArgValue();
        if (m2 == 1) {
            if (binop == 0) {
                dstvalue += srcvalue;
            } else if (binop == 1) {
                dstvalue -= srcvalue;
            } else {
                //op_illg();
                regs_pc -= 2;
                Exception(4);
                return;
            }
        } else {
            switch (binop) {
                case 0: { // BinOpAdd
                    boolean flgs = (srcvalue < 0);
                    boolean flgo = (dstvalue < 0);
                    dstvalue += srcvalue;
                    if (opsize == 1) dstvalue = (byte) dstvalue;
                    else if  (opsize == 2) dstvalue = (short) dstvalue;
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                    VFLG = (flgs && flgo && !NFLG) || (!flgs && !flgo && NFLG);
                    XFLG = CFLG = (flgs && flgo) || (!NFLG && (flgo || flgs));
                }
                break;
                case 1: { // BinOpSub
                    boolean flgs = (srcvalue < 0);
                    boolean flgo = (dstvalue < 0);
                    dstvalue -= srcvalue;
                    if (opsize == 1) dstvalue = (byte) dstvalue;
                    else if (opsize == 2) dstvalue = (short) dstvalue;
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                    VFLG = (flgs != flgo) && (NFLG != flgo);
                    XFLG = CFLG = (flgs && !flgo) || (NFLG && (!flgo || flgs));
                }
                break;
                case 3: { // BinOpSubX
                    boolean flgs = (srcvalue < 0);
                    boolean flgo = (dstvalue < 0);
                    dstvalue = dstvalue - srcvalue - (XFLG ? 1 : 0);
                    if (opsize == 1) dstvalue = (byte) dstvalue;
                    else if (opsize == 2) dstvalue = (short) dstvalue;
                    if (dstvalue != 0) ZFLG = false;
                    NFLG = (dstvalue < 0);
                    VFLG = (!flgs && flgo && !NFLG) || (flgs && !flgo && NFLG);
                    XFLG = CFLG = (flgs && !flgo) || (NFLG && (!flgo || flgs));
                }
                break;
                case 2: { // BinOpAddX
                    boolean flgs = (srcvalue < 0);
                    boolean flgo = (dstvalue < 0);
                    dstvalue += srcvalue + (XFLG ? 1 : 0);
                    if (opsize == 1) dstvalue = (byte) dstvalue;
                    else if (opsize == 2) dstvalue = (short) dstvalue;
                    if (dstvalue != 0) ZFLG = false;
                    NFLG = (dstvalue < 0);
                    XFLG = CFLG = (flgs && flgo) || (!NFLG && (flgo || flgs));
                    VFLG = (flgs && flgo && !NFLG) || (!flgs && !flgo && NFLG);
                }
                break;
                case 7: // BinOpAnd
                    dstvalue &= srcvalue;
                    VFLG = CFLG = false;
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                    break;
                case 6: // BinOpOr
                    dstvalue |= srcvalue;
                    VFLG = CFLG = false;
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                    break;
                case 8: // BinOpEor
                    dstvalue ^= srcvalue;
                    VFLG = CFLG = false;
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                    break;
                case 9: { // BinOpASL
                    int cnt = srcvalue & 63;
                    int dstvalue0 = dstvalue;
                    int comparevalue;
                    if (cnt == 0) {
                        VFLG = CFLG = false;
                    } else {
                        if (cnt > 32) {
                            dstvalue = 0;
                        } else {
                            dstvalue <<= (cnt - 1);
                        }
                        if (opsize == 1) dstvalue = (byte) dstvalue;
                        else if (opsize == 2) dstvalue = (short) dstvalue;
                        CFLG = XFLG = (dstvalue < 0);
                        dstvalue <<= 1;
                        if (opsize == 1) dstvalue = (byte) dstvalue;
                        else if (opsize == 2) dstvalue = (short) dstvalue;
                    }
                    //comparevalue = dstvalue >>> cnt;
                    if (dstvalue < 0) {
                        comparevalue = -((- dstvalue) >>> cnt);
                    } else {
                        comparevalue = dstvalue >>> cnt;
                    }
                    VFLG = (comparevalue != dstvalue0);
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                }
                break;
                case 10: { // BinOpASR
                    int cnt = srcvalue & 63;
                    NFLG = (dstvalue < 0);
                    VFLG = false;
                    if (cnt == 0) CFLG = false;
                    else {
                        if (NFLG) dstvalue = ~dstvalue;
                        if (opsize == 1) dstvalue &= 0xFF;
                        else if (opsize == 2) dstvalue &= 0xFFFF;
                        if (cnt > 32) {
                            dstvalue = 0;
                        } else {
                            dstvalue >>>= cnt - 1;
                        }
                        CFLG = (dstvalue & 1) != 0;
                        dstvalue >>>= 1;
                        if (NFLG) {
                            CFLG = !CFLG;
                            dstvalue = (~dstvalue);
                        }
                        XFLG = CFLG;
                    }
                    ZFLG = (dstvalue == 0);
                }
                break;
                case 11: { // BinOpLSL
                    int cnt = srcvalue & 63;
                    if (cnt == 0) CFLG = false;
                    else {
                        if (cnt > 32) {
                            dstvalue = 0;
                        } else {
                            dstvalue <<= cnt - 1;
                        }
                        if (opsize == 1) dstvalue = (byte) dstvalue;
                        else if (opsize == 2) dstvalue = (short) dstvalue;
                        CFLG = XFLG = (dstvalue < 0);
                        dstvalue <<= 1;
                        if (opsize == 1) dstvalue = (byte) dstvalue;
                        else if (opsize == 2) dstvalue = (short) dstvalue;
                    }
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                    VFLG = false;
                }
                break;
                case 12: { // BinOpLSR
                    int cnt = srcvalue & 63;
                    if (cnt == 0) CFLG = false;
                    else {
                        if (opsize == 1) dstvalue &= 0xFF;
                        else if (opsize == 2) dstvalue &= 0xFFFF;
                        if (cnt > 32) {
                            dstvalue = 0;
                        } else {
                            dstvalue >>>= cnt - 1;
                        }
                        CFLG = XFLG = (dstvalue & 1) != 0;
                        dstvalue >>>= 1;
                    }
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                    VFLG = false;
                }
                break;
                case 15: { // BinOpROL
                    int cnt = srcvalue & 63;
                    if (cnt == 0) CFLG = false;
                    else {
                        for (; cnt != 0; --cnt) {
                            CFLG = (dstvalue < 0);
                            dstvalue <<= 1;
                            if (CFLG) dstvalue |= 1;
                            if (opsize == 1) dstvalue = (byte) dstvalue;
                            else if (opsize == 2) dstvalue = (short) dstvalue;
                        }
                    }
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                    VFLG = false;
                }
                break;
                case 13: { // BinOpRXL
                    int cnt = srcvalue & 63;
                    if (cnt == 0) CFLG = XFLG;
                    else {
                        for (; cnt != 0; --cnt) {
                            CFLG = (dstvalue < 0);
                            dstvalue <<= 1;
                            if (XFLG) dstvalue |= 1;
                            if (opsize == 1) dstvalue = (byte) dstvalue;
                            else if (opsize == 2) dstvalue = (short) dstvalue;
                            XFLG = CFLG;
                        }
                    }
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                    VFLG = false;
                }
                break;
                case 16: { // BinOpROR
                    int cnt = srcvalue & 63;
                    int cmask = 1 << ((opsize << 3) - 1);
                    if (cnt == 0) CFLG = false;
                    else {
                        if (opsize == 1) dstvalue &= 0xFF;
                        else if (opsize == 2) dstvalue &= 0xFFFF;
                        for (; cnt != 0; --cnt) {
                            CFLG = (dstvalue & 1) != 0;
                            dstvalue >>>= 1;
                            if (CFLG) dstvalue |= cmask;
                        }
                        if (opsize == 1) dstvalue = (byte) dstvalue;
                        else if (opsize == 2) dstvalue = (short) dstvalue;
                    }
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                    VFLG = false;
                }
                break;
                case 14: { // BinOpRXR
                    int cnt = srcvalue & 63;
                    int cmask = 1 << ((opsize << 3) - 1);
                    if (cnt != 0) CFLG = XFLG;
                    else {
                        if (opsize == 1) dstvalue &= 0xFF;
                        else if (opsize == 2) dstvalue &= 0xFFFF;
                        for (; cnt != 0; --cnt) {
                            CFLG = (dstvalue & 1) != 0;
                            dstvalue >>>= 1;
                            if (XFLG) dstvalue |= cmask;
                            XFLG = CFLG;
                        }
                        if (opsize == 1) dstvalue = (byte) dstvalue;
                        else if (opsize == 2) dstvalue = (short) dstvalue;
                    }
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                    VFLG = false;
                }
                break;
                case 4: { // BinOpAddBCD
                    /* if (opsize != 1) a bug */
                    boolean flgs = (srcvalue < 0);
                    boolean flgo = (dstvalue < 0);
                    int newv_lo = (srcvalue & 0xF) + (dstvalue & 0xF) + (XFLG ? 1 : 0);
                    int newv_hi = (srcvalue & 0xF0) + (dstvalue & 0xF0);
                    int newv;
                    if (newv_lo > 9) {
                        newv_lo += 6;
                    }
                    newv = newv_hi + newv_lo;
                    CFLG = XFLG = (newv & 0x1F0) > 0x90;
                    if (CFLG) {
                        newv += 0x60;
                    }
                    dstvalue = (byte) newv;
                    if (dstvalue != 0) ZFLG = false;
                    NFLG = (dstvalue < 0);
                    VFLG = (flgs != flgo) && (NFLG != flgo);
                    /* but according to my reference book, VFLG is Undefined for ABCD */
                }
                break;
                case 5: { // BinOpSubBCD
                    boolean flgs = (srcvalue < 0);
                    boolean flgo = (dstvalue < 0);
                    int newv_lo = (dstvalue & 0xF) - (srcvalue & 0xF) - (XFLG ? 1 : 0);
                    int newv_hi = (dstvalue & 0xF0) - (srcvalue & 0xF0);
                    int newv;

                    if (newv_lo > 9) {
                        newv_lo -= 6;
                        newv_hi -= 0x10;
                    }
                    newv = newv_hi + (newv_lo & 0xF);
                    CFLG = XFLG = (newv_hi & 0x1F0) > 0x90;
                    if (CFLG) {
                        newv -= 0x60;
                    }
                    dstvalue = (byte) newv;
                    if (dstvalue != 0) ZFLG = false;
                    NFLG = (dstvalue < 0);
                    VFLG = (flgs != flgo) && (NFLG != flgo);
                    /* but according to my reference book, VFLG is Undefined for SBCD */
                }
                break;
                default:
                    //op_illg();
                    regs_pc -= 2;
                    Exception(4);
                    return;
            }
        }
        SetArgValue(dstvalue);
    }

    private void DoCompare(int m1, int r1, int m2, int r2) {
        int srcvalue;
        int dstvalue;
        DecodeModeRegister(m1, r1);
        srcvalue = GetArgValue();
        DecodeModeRegister(m2, r2);
        dstvalue = GetArgValue();
        boolean flgs = (srcvalue < 0);
        boolean flgo = (dstvalue < 0);
        dstvalue -= srcvalue;
        if (m2 != 1) {
            if (opsize == 1) {
                dstvalue = (byte) dstvalue;
            } else if (opsize == 2) {
                dstvalue = (short) dstvalue;
            }
        }
        ZFLG = (dstvalue == 0);
        NFLG = (dstvalue < 0);
        VFLG = (flgs != flgo) && (NFLG != flgo);
        CFLG = (flgs && !flgo) || (NFLG && (!flgo || flgs));
    }

    private void DoCompareA(int m1, int r1, int m2, int r2) {
        int srcvalue;
        int dstvalue;
        DecodeModeRegister(m1, r1);
        srcvalue = GetArgValue();
        if (m1 == 1) {
            switch (opsize) {
                case 1:
                    srcvalue = (byte) srcvalue;
                    break;
                case 2:
                    srcvalue = (short) srcvalue;
                    break;
                case 4:
                    break;
            }
        }
        /* if m1 != 1 then GetArgValue already has signed extended to 32 bits */
        DecodeModeRegister(m2, r2);
        dstvalue = GetArgValue();
        {
            boolean flgs = (srcvalue < 0);
            boolean flgo = (dstvalue < 0);
            dstvalue -= srcvalue;
            ZFLG = (dstvalue == 0);
            NFLG = (dstvalue < 0);
            VFLG = (flgs != flgo) && (NFLG != flgo);
            CFLG = (flgs && !flgo) || (NFLG && (!flgo || flgs));
        }
    }

    private void DoBinBitOp1(int m1, int r1, int m2, int r2, int binop) {
        int srcvalue;
        int dstvalue;
        opsize = 1;
        DecodeModeRegister(m1, r1);
        srcvalue = GetArgValue();
        if (m2 == 0) {
            opsize = 4;
        }
        DecodeModeRegister(m2, r2);
        dstvalue = GetArgValue();
        if (m2 != 1) {
            if (m2 != 0) {
                srcvalue &= 7;
            } else {
                srcvalue &= 31;
            }
            ZFLG = (dstvalue & (1 << srcvalue)) == 0;
            if (binop != 17) {
                switch (binop) {
                    case 18:
                        dstvalue ^= (1 << srcvalue);
                        break;
                    case 19:
                        dstvalue &= ~(1 << srcvalue);
                        break;
                    case 20:
                        dstvalue |= (1 << srcvalue);
                        break;
                    default:
                        //op_illg();
                        regs_pc -= 2;
                        Exception(4);
                        return;
                }
                SetArgValue(dstvalue);
            }
        }
    }

    private void DoUniOp1(int m2, int r2, int uniop) {
        int dstvalue;
        if (m2 == 1) {
            //op_illg();
            regs_pc -= 2;
            Exception(4);
        } else {
            DecodeModeRegister(m2, r2);
            dstvalue = GetArgValue();

            switch (uniop) {
                case 2: {
                    boolean flgs = (dstvalue < 0);
                    dstvalue = 0 - dstvalue - (XFLG ? 1 : 0);
                    if (opsize == 1) dstvalue = (byte) dstvalue;
                    else if (opsize == 2) dstvalue = (short) dstvalue;
                    if (dstvalue != 0) ZFLG = false;
                    NFLG = (dstvalue < 0);
                    VFLG = (flgs && NFLG);
                    XFLG = CFLG = (flgs || NFLG);
                }
                break;
                case 1: {
                    boolean flgs = (dstvalue < 0);
                    dstvalue = 0 - dstvalue;
                    if (opsize == 1) dstvalue = (byte) dstvalue;
                    else if (opsize == 2) dstvalue = (short) dstvalue;
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                    VFLG = (flgs && NFLG);
                    XFLG = CFLG = (flgs || NFLG);
                }
                break;
                case 0: {
                    dstvalue = ~dstvalue;
                    if (opsize == 1) dstvalue = (byte) dstvalue;
                    else if (opsize == 2) dstvalue = (short) dstvalue;
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                    VFLG = CFLG = false;
                }
                break;
                case 3: {
                    int newv_lo = -(dstvalue & 0xF) - (XFLG ? 1 : 0);
                    int newv_hi = -(dstvalue & 0xF0);
                    int newv;
                    if (newv_lo > 9) {
                        newv_lo -= 6;
                        newv_hi -= 0x10;
                    }
                    newv = newv_hi + (newv_lo & 0xF);
                    CFLG = XFLG = (newv_hi & 0x1F0) > 0x90;
                    if (CFLG) {
                        newv -= 0x60;
                    }
                    dstvalue = newv;
                    if (opsize == 1) dstvalue = (byte) dstvalue;
                    else if (opsize == 2) dstvalue = (short) dstvalue;
                    NFLG = (dstvalue < 0);
                    if (dstvalue != 0) ZFLG = false;
                }
                break;
                case 4: {
                    ZFLG = (dstvalue == 0);
                    NFLG = (dstvalue < 0);
                    VFLG = CFLG = false;
                    dstvalue |= 0x80;
                }
                break;
            }
            /* DoUniOp(uniop, opsize, dstvalue, CCRregister); */
            SetArgValue(dstvalue);
        }
    }
    
    private void DoBinOpMul1(int m1, int r1, int m2, int r2, int binop) {
        int srcvalue;
        int dstvalue;
        opsize = 2;
        DecodeModeRegister(m1, r1);
        srcvalue = GetArgValue();
        DecodeModeRegister(m2, r2);
        dstvalue = GetArgValue();
        if (m2 == 1) {
            //op_illg();
            regs_pc -= 2;
            Exception(4);
            return;
        } else {
            switch (binop) {
                case 21:
                    dstvalue = (dstvalue & 0xFFFF) * (srcvalue & 0xFFFF);
                    break;
                case 22:
                    dstvalue = ((short) dstvalue) * ((short) srcvalue);
                    break;
                default:
                    //op_illg();
                    regs_pc -= 2;
                    Exception(4);
                    return;
            }
            VFLG = CFLG = false;
            ZFLG = (dstvalue == 0);
            NFLG = (dstvalue < 0);
        }
        opsize = 4;
        SetArgValue(dstvalue);
    }

    private void DoBinOpDiv1(int m1, int r1, int m2, int r2, int binop) {
        int srcvalue;
        int dstvalue;

        opsize = 2;
        DecodeModeRegister(m1, r1);
        srcvalue = GetArgValue();
        opsize = 4;
        DecodeModeRegister(m2, r2);
        dstvalue = GetArgValue();
        if (m2 == 1) {
            //op_illg();
            regs_pc -= 2;
            Exception(4);
            return;
        } else if (srcvalue == 0) {
            Exception(5);
        } else {
            switch (binop) {
                case 23: { // BinOpDivU
                    long dl = dstvalue & 0x0ffffffffL;
                    srcvalue &= 0xFFFF;
                    int newv = (int) (dl / (long) srcvalue);
                    int rem = (int) (dl % srcvalue);
                    if (newv > 0xffff) {
                        VFLG = NFLG = true;
                        CFLG = false;
                    } else {
                        VFLG = CFLG = false;
                        ZFLG = ((short) newv) == 0;
                        NFLG = ((short) newv) < 0;
                        dstvalue = (newv & 0xffff) | (rem << 16);
                    }
                }
                break;
                case 24: { // BinOpDivS
                    int newv = dstvalue / (short) srcvalue;
                    int rem = (dstvalue % (short) srcvalue) & 0xFFFF;
                    if ((newv & 0xffff8000) != 0 && (newv & 0xffff8000) != 0xffff8000) {
                        VFLG = NFLG = true;
                        CFLG = false;
                    } else {
                        if (((short) rem) < 0 != dstvalue < 0) rem = -rem;
                        VFLG = CFLG = false;
                        ZFLG = ((short) newv) == 0;
                        NFLG = ((short) newv) < 0;
                        dstvalue = (newv & 0xffff) | (rem << 16);
                    }
                }
                break;
                default:
                    //op_illg();
                    regs_pc -= 2;
                    Exception(4);
                    return;
            }
        }
        SetArgValue(dstvalue);
    }

    private int opcode;
    // bitop = 17 + ((opcode >> 6) & 3);
    // octdat = x == 0 ? 8 : x;
    
    private void FindOpSizeFromb76() {
        switch ((opcode >> 6) & 3) {
            case 0:
                opsize = 1;
                break;
            case 1:
                opsize = 2;
                break;
            case 2:
                opsize = 4;
                break;
        }
    }

    private void DoCode0() {
        int mode = (opcode >> 3) & 7;
        if ((opcode  & 0x100) != 0) {
            if (mode == 1) {
                /* MoveP 0000ddd1mm001aaa */
                switch ((opcode >> 6) & 3) {
                    case 0: {
                        int srcreg = opcode & 7;
                        int dstreg = (opcode >> 9) & 7;
                        int memp = regs[srcreg + 8] + nextiword();
                        int val = (get_byte(memp) << 8) | get_byte(memp + 2);
                        regs[dstreg] = (regs[dstreg] & 0xffff0000) | val;
                    }
                    break;
                    case 1: {
                        int srcreg = opcode & 7;
                        int dstreg = (opcode >> 9) & 7;
                        int memp = regs[srcreg + 8] + nextiword();
                        regs[dstreg] = (get_byte(memp) << 24) | (get_byte(memp + 2) << 16)
                                | (get_byte(memp + 4) << 8) | get_byte(memp + 6);
                    }
                    break;
                    case 2: {
                        int srcreg = (opcode >> 9) & 7;
                        int dstreg = opcode & 7;
                        int src = regs[srcreg];
                        int memp = regs[dstreg + 8] + nextiword();
                        put_byte(memp, src >> 8);
                        put_byte(memp + 2, src);
                    }
                    break;
                    case 3: {
                        int srcreg = (opcode >> 9) & 7;
                        int dstreg = opcode & 7;
                        int src = regs[srcreg];
                        int memp = regs[dstreg + 8] + nextiword();
                        put_byte(memp, src >> 24);
                        put_byte(memp + 2, src >> 16);
                        put_byte(memp + 4, src >> 8);
                        put_byte(memp + 6, src);
                    }
                    break;
                }

            } else {
                /* dynamic bit, Opcode = 0000ddd1ttmmmrrr */
                DoBinBitOp1(0, (opcode >> 9) & 7, mode, opcode & 7, 17 + ((opcode >> 6) & 3));
            }
        } else if (((opcode >> 9) & 7) == 4) {
            /* static bit 00001010ssmmmrrr */
            DoBinBitOp1(7, 4, mode, opcode & 7, 17 + ((opcode >> 6) & 3));
        } else if (((opcode >> 9) & 7) == 6) {
            FindOpSizeFromb76();
            DoCompare(7, 4, mode, opcode & 7);
        } else if (((opcode >> 9) & 7) == 7) {
            /* MoveS */
            if (!regs_s) {
                regs_pc -= 2;
                Exception(8);
            } else {
                int dstreg = opcode & 7;
                int extra = nextiword();
                if ((extra & 0x800) != 0) {
                    int src = regs[(extra >> 12) & 15];
                    int dsta = regs[dstreg + 8];
                    put_byte(dsta, src);
                } else {
                    int srca = regs[dstreg + 8];
                    int src = get_byte(srca);
                    if ((extra & 0x8000) != 0) {
                        regs[((extra >> 12) & 7) + 8] = (byte) src;
                    } else {
                        extra = (extra >> 12) & 7;
                        regs[extra] = (regs[extra] & 0xffffFF00) | src;
                    }
                }
            }
        } else {
            int BinOp;
            FindOpSizeFromb76();
            switch ((opcode >> 9) & 7) {
                case 0:
                    BinOp = 6;
                    break;
                case 1:
                    BinOp = 7;
                    break;
                case 2:
                    BinOp = 1;
                    break;
                case 3:
                    BinOp = 0;
                    break;
                case 5:
                default:
                    /* for compiler. should be 0, 1, 2, 3, or 5 */
                    BinOp = 8;
                    break;
            }
            if (((opcode & 0x0400) == 0) && (mode == 7) && ((opcode & 7) == 4)) {
                if (((opcode >> 6) & 3) == 0) {
                    DoBinOp1(7, 4, 10, opcode & 7, BinOp);
                } else if (!regs_s) {
                    regs_pc -= 2;
                    Exception(8);
                } else {
                    DoBinOp1(7, 4, 11, opcode & 7, BinOp);
                }
            } else {
                DoBinOp1(7, 4, mode, opcode & 7, BinOp);
            }
        }
    }

    private void DoCode1() {
        opsize = 1;
        DoMove((opcode >> 3) & 7, opcode & 7, (opcode >> 6) & 7, (opcode >> 9) & 7);
    }

    private void DoCode2() {
        opsize = 4;
        DoMove((opcode >> 3) & 7, opcode & 7, (opcode >> 6) & 7, (opcode >> 9) & 7);
    }

    private void DoCode3() {
        opsize = 2;
        DoMove((opcode >> 3) & 7, opcode & 7, (opcode >> 6) & 7, (opcode >> 9) & 7);
    }

    private void DoCheck(int m1, int r1, int m2, int r2) {
        int srcvalue;
        int dstvalue;
        DecodeModeRegister(m1, r1);
        srcvalue = GetArgValue();
        DecodeModeRegister(m2, r2);
        dstvalue = GetArgValue();
        if (dstvalue < 0) {
            NFLG = true;
            Exception(6);
        } else if (dstvalue > srcvalue) {
            NFLG = false;
            Exception(6);
        }
    }

    private void DoLea(int m1, int r1, int r2) {
        DecodeModeRegister(m1, r1);
        if (ArgKind == AKMemory) {
            regs[r2 + 8] = ArgAddr;
        }
    }

    private void DoTest1(int m1, int r1) {
        DecodeModeRegister(m1, r1);
        m1 = GetArgValue();
        VFLG = CFLG = false;
        ZFLG = (m1 == 0);
        NFLG = (m1 < 0);
    }
    
    private void reglist(int direction, int m1, int r1) {
        int z;
        int p;
        int regmask;
        regmask = nextiword();
        switch (m1) {
            case 3:
                if (direction == 1) {
                    p = regs[r1 + 8];
                    if (opsize == 2) {
                        for (z = 0; z < 16; ++z) {
                            if ((regmask & (1 << (z))) != 0) {
                                regs[z] = (short) get_word(p);
                                p += 2;
                            }
                        }
                        regs[r1 + 8] = p;
                    } else {
                        for (z = 0; z < 16; ++z) {
                            if ((regmask & (1 << (z))) != 0) {
                                regs[z] = get_long(p);
                                p += 4;
                            }
                        }
                    }
                    regs[r1 + 8] = p;
                }
                break;
            case 4:
                if (direction == 0) {
                    p = regs[r1 + 8];
                    if (opsize == 2) {
                        for (z = 16; --z >= 0;) {
                            if ((regmask & (1 << (15 - z))) != 0) {
                                p -= 2;
                                put_word(p, regs[z]);
                            }
                        }
                    } else {
                        for (z = 16; --z >= 0;) {
                            if ((regmask & (1 << (15 - z))) != 0) {
                                p -= 4;
                                put_long(p, regs[z]);
                            }
                        }
                    }
                    regs[r1 + 8] = p;
                }
                break;
            default:
                DecodeModeRegister(m1, r1);
                if (ArgKind == AKMemory) {
                    p = ArgAddr;
                    if (direction == 0) {
                        if (opsize == 2) {
                            for (z = 0; z < 16; ++z) {
                                if ((regmask & (1 << (z))) != 0) {
                                    put_word(p, regs[z]);
                                    p += 2;
                                }
                            }
                        } else {
                            for (z = 0; z < 16; ++z) {
                                if ((regmask & (1 << (z))) != 0) {
                                    put_long(p, regs[z]);
                                    p += 4;
                                }
                            }
                        }
                    } else if (opsize == 2) {
                        for (z = 0; z < 16; ++z) {
                            if ((regmask & (1 << (z))) != 0) {
                                regs[z] = (short) get_word(p);
                                p += 2;
                            }
                        }
                    } else {
                        for (z = 0; z < 16; ++z) {
                            if ((regmask & (1 << (z))) != 0) {
                                regs[z] = get_long(p);
                                p += 4;
                            }
                        }
                    }
                }
                break;
        }
    }

    private void DoCode4() {
        int b76 = (opcode >> 6) & 3;
        int mode = (opcode >> 3) & 7;
        if ((opcode  & 0x100) != 0) {
            switch (b76) {
                case 0:
                case 2:
                    /* Chk 0100ddd1s0mmmrrr */
                    opsize = 4 - b76;
                    DoCheck(mode, opcode & 7, 0, (opcode >> 9) & 7);
                    break;
                case 1:
                    //op_illg();
                    regs_pc -= 2;
                    Exception(4);
                    break;
                case 3:
                    /* Lea 0100aaa111mmmrrr */
                    opsize = 4;
                    DoLea(mode, opcode & 7, (opcode >> 9) & 7);
                    break;
            }
        } else {
            switch ((opcode >> 9) & 7) {
                case 0:
                    if (b76 != 3) {
                        /* NegX 01000000ssmmmrrr */
                        FindOpSizeFromb76();
                        DoUniOp1(mode, opcode & 7, 2); // UniOpNegX
                    } else {
                        opsize = 2;
                        DoMoveNoFlags(11, 0, mode, opcode & 7);
                    }
                    break;
                case 1:
                    if (b76 != 3) {
                        /* Clr 01000010ssmmmrrr */
                        FindOpSizeFromb76();
                        DoMove(8, 0, mode, opcode & 7);
                    } else {
                        /* 0100001011mmmrrr */
                        opsize = 2;
                        DoMoveNoFlags(10, 0, mode, opcode & 7);
                    }
                    break;
                case 2:
                    if (b76 != 3) {
                        /* Neg 01000100ssmmmrrr */
                        FindOpSizeFromb76();
                        DoUniOp1(mode, opcode & 7, 1); // UniOpNeg
                    } else {
                        /* 0100010011mmmrrr */
                        opsize = 2;
                        DoMoveNoFlags(mode, opcode & 7, 10, 0);
                    }
                    break;
                case 3:
                    if (b76 != 3) {
                        /* Not 01000110ssmmmrrr */
                        FindOpSizeFromb76();
                        DoUniOp1(mode, opcode & 7, 0); // UniOpNot
                    } else {
                        /* 0100011011mmmrrr */
                        opsize = 2;
                        DoMoveNoFlags(mode, opcode & 7, 11, 0);
                    }
                    break;
                case 4:
                    switch (b76) {
                        case 0:
                            /* Nbcd 0100100000mmmrrr */
                            opsize = 1;
                            DoUniOp1(mode, opcode & 7, 3); // UniOpNbcd
                            break;
                        case 1:
                            if (mode == 0) {
                                /* Swap 0100100001000rrr */
                                int srcreg = opcode & 7;
                                int src = regs[srcreg];
                                int dst = ((src >> 16) & 0xFFFF) | ((src & 0xFFFF) << 16);
                                VFLG = CFLG = false;
                                ZFLG = (dst == 0);
                                NFLG = (dst < 0);
                                regs[srcreg] = dst;
                            } else {
                                opsize = 4;
                                DecodeModeRegister(mode, opcode & 7);
                                if (ArgKind == AKMemory) {
                                    regs[15] -= 4;
                                    put_long(regs[15], ArgAddr);
                                }
                            }
                            break;
                        case 2:
                        case 3:
                            if (mode == 0) {
                                if (b76 == 2) {
                                    /* EXT.W */
                                    int srcreg = opcode & 7;
                                    int src = regs[srcreg];
                                    int dst = (byte) src;
                                    VFLG = CFLG = false;
                                    ZFLG = dst == 0;
                                    NFLG = dst < 0;
                                    regs[srcreg] = (regs[srcreg] & 0xffff0000) | ((dst) & 0xffff);
                                } else {
                                    /* EXT.L */
                                    int srcreg = opcode & 7;
                                    int src = regs[srcreg];
                                    int dst = (short) src;
                                    VFLG = CFLG = false;
                                    ZFLG = dst == 0;
                                    NFLG = dst < 0;
                                    regs[srcreg] = dst;
                                }
                            } else {
                                opsize = 2 * b76 - 2;
                                reglist(0, mode, opcode & 7);
                            }
                            break;
                    }
                    break;
                case 5:
                    if (b76 != 3) {
                        /* Tst  01001010ssmmmrrr */
                        FindOpSizeFromb76();
                        DoTest1(mode, opcode & 7);
                    } else if ((mode == 7) && (opcode & 7) != 0) {
                        //op_illg();
                        regs_pc -= 2;
                        Exception(4);
                        /* the ILLEGAL instruction */
                    } else {
                        /* Tas 0100101011mmmrrr */
                        opsize = 1;
                        DoUniOp1(mode, opcode & 7, 4); // UniOpTAS
                    }
                    break;
                case 6:
                    if (((opcode >> 7) & 1) == 1) {
                        opsize = (b76 << 1) - 2;
                        reglist(1, mode, opcode & 7);
                    } else {
                        /* MULS/MULU/DIVS/DIVU long operations for MC68020 */
                        //op_illg();
                        regs_pc -= 2;
                        Exception(4);
                    }
                    break;
                case 7:
                    switch (b76) {
                        case 0:
                            //op_illg();
                            regs_pc -= 2;
                            Exception(4);
                            break;
                        case 1:
                            switch (mode) {
                                case 0:
                                case 1:
                                    /* Trap 010011100100vvvv */
                                    Exception((opcode & 15) + 32);
                                    break;
                                case 2: /* Link */ {
                                    int srcreg = opcode & 7;
                                    int stackp = regs[15];
                                    stackp -= 4;
                                    regs[15] = stackp;
                                    /* only matters if srcreg == 7 */
                                    put_long(stackp, regs[srcreg + 8]);
                                    regs[srcreg + 8] = stackp;
                                    regs[15] += (short) nextiword();
                                }
                                break;
                                case 3: /* Unlk */ {
                                    int srcreg = opcode & 7;
                                    if (srcreg != 7) {
                                        int src = regs[srcreg + 8];
                                        regs[srcreg + 8] = get_long(src);
                                        regs[15] = src + 4;
                                    } else {
                                        /* wouldn't expect this to happen */
                                        regs[15] = get_long(regs[15]) + 4;
                                    }
                                }
                                break;
                                case 4:
                                    /* MOVE USP 0100111001100aaa */
                                    if (!regs_s) {
                                        regs_pc -= 2;
                                        Exception(8);
                                    } else {
                                        regs_usp = regs[(opcode & 7) + 8];
                                    }
                                    break;
                                case 5:
                                    /* MOVE USP 0100111001101aaa */
                                    if (!regs_s) {
                                        regs_pc -= 2;
                                        Exception(8);
                                    } else {
                                        regs[(opcode & 7) + 8] = regs_usp;
                                    }
                                    break;
                                case 6:
                                    opsize = 0;
                                    switch (opcode & 7) {
                                        case 0:
                                            /* Reset 0100111001100000 */
                                            if (!regs_s) {
                                                regs_pc -= 2;
                                                Exception(8);
                                            } else {
                                                customreset();
                                            }
                                            break;
                                        case 1:
                                            /* Nop Opcode = 0100111001110001 */
                                            break;
                                        case 2:
                                            /* Stop 0100111001110010 */
                                            if (!regs_s) {
                                                regs_pc -= 2;
                                                Exception(8);
                                            } else {
                                                m68k_setSR((short) nextiword());
                                                Exception(4); // m68k_setstopped();
                                            }
                                            break;
                                        case 3:
                                            /* Rte 0100111001110011 */
                                            if (!regs_s) {
                                                regs_pc -= 2;
                                                Exception(8);
                                            } else {
                                                int stackp = regs[15];
                                                m68k_setSR(get_word(stackp));
                                                stackp += 2;
                                                m68k_setpc(get_long(stackp));
                                                stackp += 4;
                                                regs[15] = stackp;
                                            }
                                            break;
                                        case 4:
                                            /* Rtd 0100111001110100 */
                                            //op_illg();
                                            regs_pc -= 2;
                                            Exception(4);
                                            break;
                                        case 5:
                                            /* Rts 0100111001110101 */
                                            m68k_setpc(get_long(regs[15]));
                                            regs[15] += 4;
                                            break;
                                        case 6:
                                            /* TrapV 0100111001110110 */
                                            if (VFLG) {
                                                Exception(7);
                                            }
                                            break;
                                        case 7: /* Rtr 0100111001110111 */ {
                                            int stackp = regs[15];

                                            m68k_setCR(get_word(stackp));
                                            stackp += 2;
                                            m68k_setpc(get_long(stackp));
                                            stackp += 4;
                                            regs[15] = stackp;
                                        }
                                        break;
                                    }
                                    break;
                                case 7:
                                    //op_illg();
                                    regs_pc -= 2;
                                    Exception(4);
                                    break;
                            }
                            break;
                        case 2: /* Jsr 0100111010mmmrrr */ {
                            opsize = 0;
                            DecodeModeRegister(mode, opcode & 7);
                            if (ArgKind == AKMemory) {
                                regs[15] -= 4;
                                put_long(regs[15], regs_pc);
                                m68k_setpc(ArgAddr);
                            }
                        }
                        break;
                        case 3: /* JMP 0100111011mmmrrr */ {
                            opsize = 0;
                            DecodeModeRegister(mode, opcode & 7);
                            if (ArgKind == AKMemory) {
                                m68k_setpc(ArgAddr);
                            }
                        }
                        break;
                    }
                    break;
            }
        }
    }

    private void DoCode5() {
        int mode = (opcode >> 3) & 7;
        if (((opcode >> 6) & 3) == 3) {
            int cond = (opcode >> 8) & 15;
            if (mode == 1) {
                /* DBcc 0101cccc11001ddd */
                int srcvalue;
                int dstvalue;
                opsize = 2;
                DecodeModeRegister(7, 2);
                srcvalue = ArgAddr;
                if (!cctrue(cond)) {
                    DecodeModeRegister(0, opcode & 7);
                    dstvalue = GetArgValue();
                    --dstvalue;
                    SetArgValue(dstvalue);
                    if (dstvalue != -1) {
                        m68k_setpc(srcvalue);
                    }
                }
            } else {
                /* Scc 0101cccc11mmmrrr */
                opsize = 1;
                DecodeModeRegister(mode, opcode & 7);
                SetArgValue(cctrue(cond) ? 0xff : 0);
            }
        } else {
            int BinOp;
            /* 0101nnnossmmmrrr */
            FindOpSizeFromb76();
            if ((opcode & 0x100) == 0) {
                BinOp = 0;
                /* DivU 0101nnn0ssmmmrrr */
            } else {
                BinOp = 1;
                /* DivS 0101nnn1ssmmmrrr */
            }
            int r9 = (opcode >> 9) & 7;
            DoBinOp1(8, r9 == 0 ? 8 : r9, mode, opcode & 7, BinOp);
        }
    }

    private void DoCode6() {
        int cond = (opcode >> 8) & 15;
        int src = opcode & 255;
        int s = regs_pc;
        if (src == 0) {
            s += (short) nextiword();
        } else if (src == 255) {
            s += nextilong();
        } else {
            s += (byte) src;
        }
        if (cond == 1) {
            /* Bsr 01100001nnnnnnnn */
            regs[15] -= 4;
            put_long(regs[15], regs_pc);
            m68k_setpc(s);
        } else if (cctrue(cond)) { // Bra/Bcc 0110ccccnnnnnnnn
            m68k_setpc(s);
        }
    }

    private void DoCode7() {
        /* MoveQ 0111ddd0nnnnnnnn */
        int src = (byte) opcode;
        int dstreg = (opcode >> 9) & 7;
        VFLG = CFLG = false;
        ZFLG = src == 0;
        NFLG = src < 0;
        regs[dstreg] = src;
    }

    private void DoCode8() {
        int b8 = opcode & 0x100;
        int mode = (opcode >> 3) & 7;
        if (((opcode >> 6) & 3) == 3) {
            int BinOp;
            if (b8 == 0) {
                BinOp = 23;
                /* DivU 1000ddd011mmmrrr */
            } else {
                BinOp = 24;
                /* DivS 1000ddd111mmmrrr */
            }
            DoBinOpDiv1(mode, opcode & 7, 0, (opcode >> 9) & 7, BinOp);
        } else if ((b8 != 0) && (mode < 2)) {
            /* SBCD 1000xxx10000mxxx */
            opsize = 1;
            if (mode == 0) {
                DoBinOp1(0, opcode & 7, 0, (opcode >> 9) & 7, 5);
            } else {
                DoBinOp1(4, opcode & 7, 4, (opcode >> 9) & 7, 5);
            }
        } else {
            /* OR 1000dddmssmmmrrr */
            FindOpSizeFromb76();
            if (b8 != 0) {
                DoBinOp1(0, (opcode >> 9) & 7, mode, opcode & 7, 6);
            } else {
                DoBinOp1(mode, opcode & 7, 0, (opcode >> 9) & 7, 6);
            }
        }
    }

    private void DoCode9() {
        int mode = (opcode >> 3) & 7;
        if (((opcode >> 6) & 3) == 3) {
            /* SUBA 1t01dddm11mmmrrr */
            opsize = ((opcode >> 7) & 2) + 2;
            DoBinOp1(mode, opcode & 7, 1, (opcode >> 9) & 7, 1);
        } else {
            FindOpSizeFromb76();
            if ((opcode & 0x100) == 0) {
                DoBinOp1(mode, opcode & 7, 0, (opcode >> 9) & 7, 1);
            } else if (mode >= 2) {
                DoBinOp1(0, (opcode >> 9) & 7, mode, opcode & 7, 1);
            } else if (mode == 0) {
                DoBinOp1(0, opcode & 7, 0, (opcode >> 9) & 7, 3);
            } else {
                DoBinOp1(4, opcode & 7, 4, (opcode >> 9) & 7, 3);
            }
        }
    }

    private void DoCodeA() {
        regs_pc -= 2;
        Exception(0xA);
    }

    private void DoCodeB() {
        int mode = (opcode >> 3) & 7;
        if (((opcode >> 6) & 3) == 3) {
            opsize = ((opcode >> 7) & 2) + 2;
            DoCompareA(mode, opcode & 7, 1, (opcode >> 9) & 7);
        } else if (((opcode & 0x100) != 0) && (mode == 1)) {
            /* CmpM 1011ddd1ss001rrr */
            FindOpSizeFromb76();
            DoCompare(3, opcode & 7, 3, (opcode >> 9) & 7);
        } else if ((opcode & 0x100) != 0) {
            /* Eor 1011ddd1ssmmmrrr */
            FindOpSizeFromb76();
            DoBinOp1(0, (opcode >> 9) & 7, mode, opcode & 7, 8);
        } else {
            /* Cmp 1011ddd0ssmmmrrr */
            FindOpSizeFromb76();
            DoCompare(mode, opcode & 7, 0, (opcode >> 9) & 7);
        }
    }

    private void DoCodeC() {
        int mode = (opcode >> 3) & 7;
        int b76 = (opcode >> 6) & 3;
        if (b76 == 3) {
            int BinOp;
            if ((opcode & 0x100) == 0) {
                BinOp = 21;
                /* MulU 1100ddd011mmmrrr */
            } else {
                BinOp = 22;
                /* MulS 1100ddd111mmmrrr */
            }
            DoBinOpMul1(mode, opcode & 7, 0, (opcode >> 9) & 7, BinOp);
        } else if ((mode < 2) && ((opcode & 0x100) != 0)) {
            switch (b76) {
                case 0:
                    /* ABCD 1100ddd10000mrrr */
 /* does anyone use this? */
                    opsize = 1;
                    if (mode == 0) {
                        DoBinOp1(0, opcode & 7, 0, (opcode >> 9) & 7, 4);
                    } else {
                        DoBinOp1(4, opcode & 7, 4, (opcode >> 9) & 7, 4);
                    }
                    break;
                case 1:
                    /* Exg 1100ddd10100trrr, opsize = 4 */
                    if (mode == 0) {
                        int srcreg = (opcode >> 9) & 7;
                        int dstreg = opcode & 7;
                        int src = regs[srcreg];
                        int dst = regs[dstreg];
                        regs[srcreg] = dst;
                        regs[dstreg] = src;
                    } else {
                        int srcreg = (opcode >> 9) & 7;
                        int dstreg = opcode & 7;
                        int src = regs[srcreg + 8];
                        int dst = regs[dstreg + 8];
                        regs[srcreg + 8] = dst;
                        regs[dstreg + 8] = src;
                    }
                    break;
                case 2: {
                    /* Exg 1100ddd110001rrr, opsize = 4 */
                    int srcreg = (opcode >> 9) & 7;
                    int dstreg = opcode & 7;
                    int src = regs[srcreg];
                    int dst = regs[dstreg +8];
                    regs[srcreg] = dst;
                    regs[dstreg + 8] = src;
                }
                break;
            }
        } else {
            /* And 1100dddmssmmmrrr */
            FindOpSizeFromb76();
            if ((opcode & 0x100) != 0) {
                DoBinOp1(0, (opcode >> 9) & 7, mode, opcode & 7, 7);
            } else {
                DoBinOp1(mode, opcode & 7, 0, (opcode >> 9) & 7, 7);
            }
        }
    }

    private void DoCodeD() {
        int mode = (opcode >> 3) & 7;
        if (((opcode >> 6) & 3) == 3) {
            /* ADDA 1t01dddm11mmmrrr */
            opsize = ((opcode >> 7) & 2) + 2;
            DoBinOp1(mode, opcode & 7, 1, (opcode >> 9) & 7, 0);
        } else {
            FindOpSizeFromb76();
            if ((opcode & 0x100) == 0) {
                DoBinOp1(mode, opcode & 7, 0, (opcode >> 9) & 7, 0);
            } else if (mode >= 2) {
                DoBinOp1(0, (opcode >> 9) & 7, mode, opcode & 7, 0);
            } else if (mode == 0) {
                DoBinOp1(0, opcode & 7, 0, (opcode >> 9) & 7, 2);
            } else {
                DoBinOp1(4, opcode & 7, 4, (opcode >> 9) & 7, 2);
            }
        }
    }

    private int rolops(int x) {
        int binop;
        switch (x) {
            case 0:
                binop = 9; // BinOpASL
                break;
            case 1:
                binop = 11; // BinOpLSL
                break;
            case 2:
                binop = 13; // BinOpRXL
                break;
            case 3:
            default:
                /* for compiler. should be 0, 1, 2, or 3 */
                binop = 15; // BinOpROL
                break;
        }
        if ((opcode & 0x100) == 0) {
            binop++; /* 'R' */
        } /* else 'L' */
        return binop;
    }

    private void DoCodeE() {
        int mode = (opcode >> 3) & 7;
        if (((opcode >> 6) & 3) == 3) {
            opsize = 2;
            DoBinOp1(8, 1, mode, opcode & 7, rolops((opcode >> 9) & 7));
        } else {
            FindOpSizeFromb76();
            if (mode < 4) {
                /* 1110cccdss0ttddd */
                int r9 = (opcode >> 9) & 7;
                DoBinOp1(8, r9 == 0 ? 8 : r9, 0, opcode & 7, rolops(mode & 3));
            } else {
                /* 1110rrrdss1ttddd */
                DoBinOp1(0, (opcode >> 9) & 7, 0, opcode & 7, rolops(mode & 3));
            }
        }
    }

    private void DoCodeF() {
        //op_illg();
        regs_pc -= 2;
        Exception(4);
    }

    private int MaxInstructionsToGo;

    private void m68k_go_MaxInstructions() {
        /* MaxInstructionsToGo must be >= 1 on entry */
        do {
            opcode = nextiword();
            switch (opcode >> 12) {
                case 0x0:
                    DoCode0();
                    break;
                case 0x1:
                    DoCode1();
                    break;
                case 0x2:
                    DoCode2();
                    break;
                case 0x3:
                    DoCode3();
                    break;
                case 0x4:
                    DoCode4();
                    break;
                case 0x5:
                    DoCode5();
                    break;
                case 0x6:
                    DoCode6();
                    break;
                case 0x7:
                    DoCode7();
                    break;
                case 0x8:
                    DoCode8();
                    break;
                case 0x9:
                    DoCode9();
                    break;
                case 0xA:
                    DoCodeA();
                    break;
                case 0xB:
                    DoCodeB();
                    break;
                case 0xC:
                    DoCodeC();
                    break;
                case 0xD:
                    DoCodeD();
                    break;
                case 0xE:
                    DoCodeE();
                    break;
                case 0xF:
                    DoCodeF();
                    break;
            }
            //System.out.println("pc = " + Integer.toHexString(regs_pc));
        } while (--MaxInstructionsToGo != 0);
    }

    private int MoreInstructionsToGo;

    private void NeedToGetOut() {
        if (MaxInstructionsToGo != 0) {
            MoreInstructionsToGo += (MaxInstructionsToGo - 1);
            /* not counting the current instruction */
            MaxInstructionsToGo = 1;
        }
    }

    private void do_trace() {
        regs_TracePending = true;
        NeedToGetOut();
    }

    private void SetAutoVector() {
        regs_AutoVectorPending = true;
        NeedToGetOut();
    }

    private void SetExternalInterruptPending() {
        regs_ExternalInterruptPending = true;
        NeedToGetOut();
    }

    private void m68k_go_nInstructions(int n) {
        MaxInstructionsToGo = n;
        do {
            MoreInstructionsToGo = 0;
            if (regs_t1) {
                do_trace();
            }
            m68k_go_MaxInstructions();
            if (regs_AutoVectorPending) {
                Exception(regs_intmask + 24);
                regs_AutoVectorPending = false;
            }
            if (regs_ExternalInterruptPending) {
                DoCheckExternalInterruptPending();
                regs_ExternalInterruptPending = false;
            }
            if (regs_TracePending) {
                Exception(9);
            }
            MaxInstructionsToGo = MoreInstructionsToGo;
        } while (MaxInstructionsToGo != 0);
    }
    // Motorola 68K эмулятор конец
 
    private void SixtiethSecondNotify() {
        VIA_Timer();
        if (!vDraw) repaint();
        Mouse_Update2();
        KeyBoard_Update();

        VIA_Int_Vertical_Blanking();
        Sony_Update();
    }

    private void OneSecondNotify() {
        RTC_Interrupt();
        VIA_Int_One_Second();
    }

    private int TicksLeftInSecond = 60;

    private void DoEach60th() {
        SixtiethSecondNotify();
	if (--TicksLeftInSecond <= 0) {
            OneSecondNotify();
            TicksLeftInSecond = 60;
	}
    }

    @Override
    public void run() {
        long LastTime = System.currentTimeMillis();
        do {
            m68k_go_nInstructions(1024);
            long newTime = System.currentTimeMillis();
            if (newTime - LastTime > 16) {
                LastTime = newTime;
                DoEach60th();
            }
        } while (true);
    }

    public static void main(String[] args) {
        Emul emul = new Emul();
        emul.addWindowListener(new WindowAdapter(){
            @Override
            public void windowClosing(WindowEvent ev){
                System.exit(0);
            }
        });
        emul.start();
        if (args.length != 0) {
            for (int i = 0; i < args.length; i ++) {
                if (!emul.Sony_Insert0(new File(args[i]))) {
                    System.err.println("Cannot mount disk image!");
                }
            }
        }
    }

    @Override
    public void mouseClicked(MouseEvent e) {
        CurMouseButton = true;
    }

    @Override
    public void mousePressed(MouseEvent e) {
        CurMouseButton = true;
    }

    @Override
    public void mouseReleased(MouseEvent e) {
        CurMouseButton = false;
    }

    @Override
    public void mouseEntered(MouseEvent e) {}

    @Override
    public void mouseExited(MouseEvent e) {}

    private void setMouse(int x, int y) {
        x = (int) (x * ((float) SCR_WIDTH / (getWidth() - 24)));
        y = (int) (y * ((float) SCR_HEIGHT / (getHeight() - 46)));
        //
        if (x < 0) x = 0; else
        if (x >= SCR_WIDTH) x = SCR_WIDTH - 1;
        if (y < 0) y = 0; else
        if (y >= SCR_HEIGHT) y = SCR_HEIGHT - 1;
        CurMouseH = x;
        CurMouseV = y;
    }

    @Override
    public void mouseDragged(MouseEvent e) {
        setMouse(e.getX() - vX, e.getY() - vY);
    }

    @Override
    public void mouseMoved(MouseEvent e) {
        setMouse(e.getX() - vX, e.getY() - vY);
    }

    @Override
    public void keyTyped(KeyEvent e) {}

    // Map Java key code to MAC scan code
    private byte[] scantbl;

    // Create and initialize key translator
    void InitKeyboardMap()
    {
        scantbl = new byte[256];
        scantbl[KeyEvent.VK_ESCAPE] = 0x35;
        scantbl[KeyEvent.VK_1] = 0x12;
        scantbl[KeyEvent.VK_2] = 0x13;
        scantbl[KeyEvent.VK_3] = 0x14;
        scantbl[KeyEvent.VK_4] = 0x15;
        scantbl[KeyEvent.VK_5] = 0x17;
        scantbl[KeyEvent.VK_6] = 0x16;
        scantbl[KeyEvent.VK_7] = 0x1A;
        scantbl[KeyEvent.VK_8] = 0x1C;
       	scantbl[KeyEvent.VK_9] = 0x19;
        scantbl[KeyEvent.VK_0] = 0x1D;
        scantbl[KeyEvent.VK_MINUS] = 0x1B;
        scantbl[KeyEvent.VK_EQUALS] = 0x18;
        scantbl[KeyEvent.VK_BACK_SPACE] = 0x33;
        scantbl[KeyEvent.VK_TAB] = 0x30;
        scantbl[KeyEvent.VK_Q] = 0x0C;
        scantbl[KeyEvent.VK_W] = 0x0D;
        scantbl[KeyEvent.VK_E] = 0x0E;
        scantbl[KeyEvent.VK_R] = 0x0F;
        scantbl[KeyEvent.VK_T] = 0x11;
        scantbl[KeyEvent.VK_Y] = 0x10;
        scantbl[KeyEvent.VK_U] = 0x20;
        scantbl[KeyEvent.VK_I] = 0x22;
        scantbl[KeyEvent.VK_O] = 0x1F;
        scantbl[KeyEvent.VK_P] = 0x23;
        scantbl[KeyEvent.VK_OPEN_BRACKET] = 0x21;
        scantbl[KeyEvent.VK_CLOSE_BRACKET] = 0x1E;
        scantbl[KeyEvent.VK_ENTER] = 0x24;
        scantbl[KeyEvent.VK_CONTROL] = 0x37;
        scantbl[KeyEvent.VK_A] = 0x00;
        scantbl[KeyEvent.VK_S] = 0x01;
        scantbl[KeyEvent.VK_D] = 0x02;
        scantbl[KeyEvent.VK_F] = 0x03;
        scantbl[KeyEvent.VK_G] = 0x05;
        scantbl[KeyEvent.VK_H] = 0x04;
        scantbl[KeyEvent.VK_J] = 0x26;
        scantbl[KeyEvent.VK_K] = 0x28;
        scantbl[KeyEvent.VK_L] = 0x25;
        scantbl[KeyEvent.VK_SEMICOLON] = 0x29;
        scantbl[KeyEvent.VK_QUOTE] = 0x27;
        scantbl[KeyEvent.VK_SHIFT] = 0x38; // left shift
        scantbl[KeyEvent.VK_BACK_SLASH] = 0x2A;
        scantbl[KeyEvent.VK_Z] = 0x06;
        scantbl[KeyEvent.VK_X] = 0x07;
        scantbl[KeyEvent.VK_C] = 0x08;
        scantbl[KeyEvent.VK_V] = 0x09;
        scantbl[KeyEvent.VK_B] = 0x0B;
        scantbl[KeyEvent.VK_N] = 0x2D;
        scantbl[KeyEvent.VK_M] = 0x2E;
        scantbl[KeyEvent.VK_COMMA] = 0x2B;
        scantbl[KeyEvent.VK_PERIOD] = 0x2F;
        scantbl[KeyEvent.VK_SLASH] = 0x2C;
        scantbl[KeyEvent.VK_MULTIPLY] = 0x43;
        scantbl[KeyEvent.VK_ALT] = 0x3A;
        scantbl[KeyEvent.VK_SPACE] = 0x31;
        scantbl[KeyEvent.VK_CAPS_LOCK] = 0x39;
        scantbl[KeyEvent.VK_F1] = 0x7a;
        scantbl[KeyEvent.VK_F2] = 0x78;
        scantbl[KeyEvent.VK_F3] = 0x63;
        scantbl[KeyEvent.VK_F4] = 0x76;
        scantbl[KeyEvent.VK_F5] = 0x60;
        scantbl[KeyEvent.VK_F6] = 0x61;
        scantbl[KeyEvent.VK_F7] = 0x62;
        scantbl[KeyEvent.VK_F8] = 0x64;
        scantbl[KeyEvent.VK_F9] = 0x65;
        scantbl[KeyEvent.VK_F10] = 0x6d;
        scantbl[KeyEvent.VK_NUMPAD7] = 0x59;
        scantbl[KeyEvent.VK_NUMPAD8] = 0x5B;
        scantbl[KeyEvent.VK_UP] = 0x4D;
        scantbl[KeyEvent.VK_NUMPAD9] = 0x5C;
        scantbl[KeyEvent.VK_SUBTRACT] = 0x4E;
        scantbl[KeyEvent.VK_NUMPAD4] = 0x56;
        scantbl[KeyEvent.VK_LEFT] = 0x46;
        scantbl[KeyEvent.VK_NUMPAD5] = 0x57;
        scantbl[KeyEvent.VK_NUMPAD6] = 0x58;
        scantbl[KeyEvent.VK_RIGHT] = 0x42;
        scantbl[KeyEvent.VK_ADD] = 0x45;
        scantbl[KeyEvent.VK_NUMPAD1] = 0x53;
        scantbl[KeyEvent.VK_NUMPAD2] = 0x54;
        scantbl[KeyEvent.VK_DOWN] = 0x48;
        scantbl[KeyEvent.VK_NUMPAD3] = 0x55;
        scantbl[KeyEvent.VK_NUMPAD0] = 0x52;
        scantbl[KeyEvent.VK_DECIMAL] = 0x41;
        scantbl[KeyEvent.VK_F11] = 0x67;
        scantbl[KeyEvent.VK_F12] = 0x6f;
    }

    
    @Override
    public void keyPressed(KeyEvent e) {
        //System.out.println("Key preessed: " + KeyEvent.getKeyText(e.getKeyCode()));
        Keyboard_UpdateKeyMap(scantbl[e.getKeyCode()], true);
    }

    @Override
    public void keyReleased(KeyEvent e) {
        if (e.isControlDown() && e.getKeyCode() == KeyEvent.VK_O) {
            if (fileManager.showOpenDialog(this) == JFileChooser.APPROVE_OPTION) {
                if (!Sony_Insert0(fileManager.getSelectedFile())) {
                    System.err.println("Cannot mount disk image!");
                }
            }
        } else {
            //System.out.println("Key realesed: " + KeyEvent.getKeyText(e.getKeyCode()));
            Keyboard_UpdateKeyMap(scantbl[e.getKeyCode()], false);
        }
    }
    
}
